import enum
import logging
import re
import shlex
from asyncio import create_task, get_running_loop, run_coroutine_threadsafe, TimeoutError
from discord import FFmpegPCMAudio, PCMVolumeTransformer, Embed, ClientException
from math import floor, ceil
from functools import partial
from os import remove as remove_file
from random import randint
from time import monotonic as time
from youtube_dl import YoutubeDL
from youtube_dl.utils import YoutubeDLError
from red_star.command_dispatcher import Command
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import UserPermissionError, CommandSyntaxError
from red_star.rs_utils import respond, split_message, get_guild_config, find_user, is_positive


class MusicPlayer(BasePlugin):
    name = "music_player"
    version = "2.2"
    author = "medeor413 (original by GTG3000)"
    description = "A plugin for playing audio from videos in a voice channel."
    default_config = {
        "save_audio": True,
        "video_cache_clear_age": 259200,
        "default_volume": 15,
        "youtube_dl_config": {
            "quiet": True,
            "restrictfilenames": True,
            "source_address": "0.0.0.0",
            "audioformat": "mp3",
            "default_search": "auto",
            "extractaudio": True,
            "logtostderr": False,
            "nocheckcertificate": True,
            "format": "bestaudio/best",
            "noplaylist": True,
            "no_color": True
        },
        "default": {
            "max_queue_length": 30,
            "max_video_length": 1800,
            "vote_skip_threshold": 0.5,
            "print_queue_on_edit": True,
            "idle_disconnect_time": 300
        }
    }

    async def activate(self):
        self.storage = self.config_manager.get_plugin_config_file("music_player.json")

        self.players = {}

        self.ydl_options = self.plugin_config["youtube_dl_config"].copy()
        self.ydl_options["logger"] = logging.getLogger("red_star.plugin.music_player.youtube-dl")
        self.ydl_options["extract_flat"] = "in_playlist"
        self.ydl_options["outtmpl"] = str(self.client.storage_dir / "music_cache" /
                                          self.ydl_options.get("outtmpl", "%(id)s-%(extractor)s.%(ext)s"))

    async def deactivate(self):
        for player in self.players.values():
            await player.voice_client.disconnect()

    # Command functions

    @Command("JoinVoice", "JoinVC",
             doc="Tells the bot to join the voice channel you're currently in.",
             category="music_player")
    async def _join_voice(self, msg):
        try:
            voice_channel = msg.author.voice.channel
        except AttributeError:
            raise UserPermissionError("**ANALYSIS: User is not connected to voice channel.**")
        try:
            if msg.guild.id in self.players:
                player = self.players[msg.guild.id]
                await player.voice_client.move_to(voice_channel)
            else:
                player = await self.create_player(voice_channel, msg.channel)
        except TimeoutError:
            raise CommandSyntaxError("Bot failed to join channel. Make sure bot can access the channel.")
        await respond(msg, f"**AFFIRMATIVE. Connected to channel {voice_channel.name}.**")
        return player

    @Command("LeaveVoice", "LeaveVC",
             doc="Tells the bot to leave the voice channel it's currently in.",
             category="music_player")
    async def _leave_voice(self, msg):
        player = self.get_guild_player(msg)
        if not player:
            try:
                msg.guild.voice_client.stop()
                await msg.guild.voice_client.disconnect(force=True)
                await respond(msg, "**ANALYSIS: Disconnected from voice channel.**")
                return
            except AttributeError:
                await respond(msg, "**ANALYSIS: Bot is not currently in voice channel.**")
                return
        if player.is_playing \
                and not player.voice_client.channel.permissions_for(msg.author).mute_members\
                and not self.config_manager.is_maintainer(msg.author):
            raise UserPermissionError
        player.queue.clear()
        player.already_played.clear()
        player.stop()
        await player.voice_client.disconnect()
        del self.players[msg.guild.id]
        await respond(msg, "**ANALYSIS: Disconnected from voice channel.**")

    @Command("PlaySong", "Play",
             doc="Tells the bot to queue video links for playing. User must be in the same channel as the bot. "
                 "Supports an arbitrary number of links in a single command.\n"
                 "Supports playlists. Additionally, subsets of playlists can be played using slice notation.\n"
                 "Subset syntax: http://example.url{1,3-5,-2,6-}. {1} selects the first video, {3-5} selects the "
                 "third through fifth videos, {-2} selects all videos up to the second, {6-} selects the sixth video "
                 "and all after it. Multiple slices can be used on one playlist by separating them with ,.",
             syntax="(url) [url_2] [url_3]...",
             category="music_player")
    async def _play_song(self, msg):
        player = self.get_guild_player(msg)
        if not player:
            player = await self._join_voice(msg)
        self.check_user_permission(msg.author, player)
        try:
            urls = shlex.split(msg.clean_content)[1:]
        except ValueError as e:
            raise CommandSyntaxError(e)
        await player.prepare_playlist(urls)

    @Command("SongQueue", "Queue",
             doc="Tells the bot to list the current song queue.",
             category="music_player")
    async def _list_queue(self, msg):
        player = self.get_guild_player(msg)
        if not player:
            await respond(msg, "**ANALYSIS: Bot is not in a voice channel.**")
            return
        elif player.queue and not player.current_song:
            await respond(msg, "**ANALYSIS: The queue is empty.**")
            return
        if player.current_song:
            await respond(msg, f"**ANALYSIS: Now playing:**\n```{player.print_now_playing()}```")
        if player.queue:
            queue_duration = pretty_duration(player.queue_duration)
            for split_msg in split_message(f"**ANALYSIS: Current queue: ({queue_duration})**{player.print_queue()}"):
                await respond(msg, split_msg)

    @Command("NowPlaying", "SongInfo",
             doc="Displays detailed information about the currently playing song.",
             category="music_player")
    async def _song_info(self, msg):
        player = self.get_guild_player(msg)
        if not player or not player.current_song:
            await respond(msg, "**ANALYSIS: There is no song currently playing.**")
            return
        vid = player.current_song
        desc = vid.get("description", "*No description.*")
        if len(desc) > 2048:
            desc = desc[:2045] + "..."
        embed = Embed(title=vid.get("title", "Unknown"), description=desc, url=vid["url"])
        if vid.get("thumbnail") is not None:
            embed.set_thumbnail(url=vid["thumbnail"])
        if vid.get("uploader") is not None:
            embed.set_author(name=vid["uploader"])
        if vid.get("tags") is not None:
            embed.set_footer(text=f"Tags: {', '.join(vid['tags'])}")
        play_time, duration, _ = player.progress
        rating_field = f"Views: {vid.get('view_count', 'Unknown'):,}."
        if vid.get("like_count") is not None:
            rating_field += f" {vid['like_count']:,}👍"
        if vid.get("dislike_count") is not None:
            rating_field += f"/{vid['dislike_count']:,}👎"
        if vid.get("average_rating") is not None:
            rating_field += f" (Average rating: {vid['average_rating']:.2f})"
        embed.add_field(name="Ratings", value=rating_field)
        await respond(msg, embed=embed)

    @Command("DeleteSong", "DelSong", "RMSong",
             doc="Removes a song from the bot's queue.",
             syntax="(index)",
             perms="mute_members",
             category="music_player")
    async def _delete_song(self, msg):
        player = self.get_guild_player(msg)
        if not player:
            await respond(msg, "**ANALYSIS: Bot is not in a voice channel.**")
            return
        elif not player.queue:
            await respond(msg, "**ANALYSIS: The queue is empty.**")
            return
        try:
            index = int(msg.clean_content.split(None, 1)[1])
        except IndexError:
            raise CommandSyntaxError("No arguments provided")
        except ValueError:
            raise CommandSyntaxError(f"{msg.clean_content.split(None, 1)[1]} is not a valid integer.")
        if index == 0:
            raise CommandSyntaxError("Use Skip to skip the currently playing song")
        elif index < 0:
            raise CommandSyntaxError("Provided integer must be positive")
        index -= 1
        try:
            del_song = player.queue.pop(index)
            if index in player.already_played:
                player.already_played.remove(index)
            new_already_played = set()
            for i in player.already_played:
                if i > index:
                    i -= 1
                new_already_played.add(i)
            player.already_played = new_already_played
        except IndexError:
            raise CommandSyntaxError("Integer provided is not a valid index")
        await respond(msg, f"**AFFIRMATIVE. Deleted song at position {index + 1} ({del_song['title']}).**")
        if get_guild_config(self, str(msg.guild.id), "print_queue_on_edit") and player.queue:
            for split_msg in split_message(f"**ANALYSIS: Current queue:**{player.print_queue()}"):
                await respond(msg, split_msg)

    @Command("SongVolume", "Volume",
             doc="Sets the volume at which the player plays music, between 0 and 100.\nUse --temporary flag to "
                 "automatically reset volume at the end of the current song.",
             syntax="[-t/--temporary] (0-100)",
             category="music_player")
    async def _set_volume(self, msg):
        player = self.get_guild_player(msg)
        self.check_user_permission(msg.author, player)
        new_volume = 0
        try:
            temp_flag = False
            new_volume = msg.clean_content.split(None, 2)[1:]
            if len(new_volume) > 1:
                if new_volume[0] in ("-t", "--temporary"):
                    temp_flag = True
                    new_volume = int(new_volume[1])
                else:
                    raise CommandSyntaxError
            else:
                new_volume = int(new_volume[0])
            if new_volume == 0:
                raise CommandSyntaxError("Please use PauseSong to mute the bot")
            elif not 0 < new_volume <= 100:
                raise CommandSyntaxError("Expected value between 0 and 100")
        except IndexError:
            await respond(msg, f"**ANALYSIS: Current volume is {int(player.volume * 100)}%.**")
            return
        except ValueError:
            raise CommandSyntaxError(f"Value {new_volume} is not a valid integer")
        if temp_flag:
            player.prev_volume = player.volume
        player.volume = new_volume / 100
        await respond(msg, f"**AFFIRMATIVE. Set volume to {new_volume}%.**")

    @Command("SongMode",
             doc="Tells the bot in what order the bot should play its queue, or prints the current mode if no mode "
                 "is specified.",
             syntax="[n/normal|rs/repeat_song|rq/repeat_queue|s/shuffle|sr/shuffle_repeat]",
             category="music_player")
    async def _song_mode(self, msg):
        player = self.get_guild_player(msg)
        self.check_user_permission(msg.author, player)
        try:
            arg = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            await respond(msg, f"**ANALYSIS: Current song mode is {player.song_mode}.**")
            return
        if arg in ("n", "normal", "reset"):
            player.song_mode = SongMode.NORMAL
        elif arg in ("rs", "repeat_song", "loop"):
            player.song_mode = SongMode.REPEAT_SONG
        elif arg in ("rq", "repeat", "repeat_queue", "cycle"):
            player.song_mode = SongMode.REPEAT_QUEUE
        elif arg in ("s", "shuffle", "random"):
            player.song_mode = SongMode.SHUFFLE
        elif arg in ("sr", "shuffle_repeat", "random_repeat"):
            player.song_mode = SongMode.SHUFFLE_REPEAT
            player.already_played.clear()
        else:
            raise CommandSyntaxError(f"Argument {arg} is not a valid mode")
        await respond(msg, f"**AFFIRMATIVE. Song mode changed to {player.song_mode}.**")

    @Command("Shuffle",
             doc="Enables or disables shuffling of song order.",
             syntax="[true/yes/on/false/no/off",
             category="music_player")
    async def _shuffle_toggle(self, msg):
        player = self.get_guild_player(msg)
        self.check_user_permission(msg.author, player)
        shuffle_on = player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT)
        try:
            arg = is_positive(msg.clean_content.split(None, 1)[1])
        except IndexError:
            await respond(msg, f"**ANALYSIS: Shuffle is currently {'enabled' if shuffle_on else 'disabled'}.**")
            return
        if arg:
            if shuffle_on:
                await respond(msg, f"**ANALYSIS: Shuffle is already enabled.**")
            else:
                if player.song_mode is SongMode.NORMAL:
                    player.song_mode = SongMode.SHUFFLE
                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled.**")
                elif player.song_mode is SongMode.REPEAT_QUEUE:
                    player.song_mode = SongMode.SHUFFLE_REPEAT
                    await respond(msg, f"**AFFIRMATIVE. Shuffle enabled with repeat.**")
                elif player.song_mode is SongMode.REPEAT_SONG:
                    await respond(msg, "**NEGATIVE. Shuffle cannot be enabled at the same time as song repeat.**")
        else:
            if shuffle_on:
                if player.song_mode is SongMode.SHUFFLE:
                    player.song_mode = SongMode.NORMAL
                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled.**")
                elif player.song_mode is SongMode.SHUFFLE_REPEAT:
                    player.already_played.clear()
                    player.song_mode = SongMode.REPEAT_QUEUE
                    await respond(msg, f"**AFFIRMATIVE. Shuffle disabled. Queue repeat still enabled.**")
            else:
                await respond(msg, "**ANALYSIS: Shuffle is not enabled.**")

    @Command("Repeat",
             doc="Modifies repeating mode of song queue.",
             syntax="[song/one/queue/all/false/no/off]",
             category="music_player")
    async def _repeat_mode(self, msg):
        player = self.get_guild_player(msg)
        self.check_user_permission(msg.author, player)
        try:
            arg = msg.clean_content.split(None, 1)[1].lower()
            if arg not in ("song", "one", "queue", "all"):
                arg = is_positive(arg)
                if arg:
                    raise CommandSyntaxError
        except IndexError:
            if player.song_mode is SongMode.REPEAT_SONG:
                repeat_mode = "set to repeat current song"
            elif player.song_mode is SongMode.REPEAT_QUEUE:
                repeat_mode = "set to repeat the queue"
            elif player.song_mode is SongMode.SHUFFLE_REPEAT:
                repeat_mode = "set to repeat the queue. Shuffle is also enabled"
            else:
                repeat_mode = "disabled"
            await respond(msg, f"**ANALYSIS: Repeat is currently {repeat_mode}.**")
            return
        if arg in ("song", "one"):
            if player.song_mode is SongMode.REPEAT_SONG:
                await respond(msg, "**ANALYSIS: Song repeat is already enabled.**")
            elif player.song_mode in (SongMode.SHUFFLE, SongMode.SHUFFLE_REPEAT):
                await respond(msg, "**NEGATIVE. Song repeat cannot be enabled at the same time as shuffle.**")
            else:
                player.song_mode = SongMode.REPEAT_SONG
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to song.**")
        elif arg in ("queue", "all"):
            if player.song_mode in (SongMode.REPEAT_QUEUE, SongMode.SHUFFLE_REPEAT):
                await respond(msg, "**ANALYSIS: Queue repeat is already enabled.**")
            elif player.song_mode is SongMode.SHUFFLE:
                player.song_mode = SongMode.SHUFFLE_REPEAT
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue with shuffle.**")
            else:
                player.song_mode = SongMode.REPEAT_QUEUE
                await respond(msg, "**AFFIRMATIVE. Repeat mode set to queue.**")
        else:
            if player.song_mode in (SongMode.NORMAL, SongMode.SHUFFLE):
                await respond("**ANALYSIS: Repeat mode is disabled.**")
            elif player.song_mode is SongMode.SHUFFLE_REPEAT:
                player.already_played.clear()
                player.song_mode = SongMode.SHUFFLE
                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled. Shuffle still enabled.**")
            else:
                player.song_mode = SongMode.NORMAL
                await respond(msg, "**AFFIRMATIVE. Repeat mode disabled.**")

    @Command("SkipSong", "Skip",
             doc="Tells the bot to skip the currently playing song.",
             category="music_player")
    async def _skip_song(self, msg):
        player = self.get_guild_player(msg)
        self.check_user_permission(msg.author, player)
        if not player or not player.is_playing:
            await respond(msg, "**ANALYSIS: No music currently playing.**")
            return
        elif player.voice_client.channel.permissions_for(msg.author).mute_members\
                or self.config_manager.is_maintainer(msg.author):
            await respond(msg, "**ANALYSIS: Skipping to next song in queue...**")
            player.stop()
        else:
            await player.skip_vote(msg.author.id)

    @Command("PauseSong", "Pause", "ResumeSong", "Resume",
             doc="Tells the bot to pause or resume the current song.",
             perms="mute_members",
             category="music_player")
    async def _pause_song(self, msg):
        player = self.get_guild_player(msg)
        if not player or player.is_playing is False:
            await respond(msg, "**WARNING: No music currently playing.**")
            return
        if player.toggle_pause():
            await respond(msg, "**ANALYSIS: Song paused.**")
        else:
            await respond(msg, "**ANALYSIS: Song resumed.**")

    @Command("StopMusic", "StopSong", "Stop",
             doc="Tells the bot to stop playing songs and empty the queue.",
             perms="mute_members",
             category="music_player")
    async def _stop_music(self, msg):
        player = self.get_guild_player(msg)
        if not player or not player.is_playing:
            await respond(msg, "**ANALYSIS: No music currently playing.")
            return
        player.queue.clear()
        player.already_played.clear()
        player.stop()
        await respond(msg, "**ANALYSIS: The music has been stopped and the queue has been cleared.**")

    @Command("MusicBan",
             doc="Bans or unbans a user from using music bot commands, or lists bans if no user specified..",
             syntax="(user)",
             perms="mute_members",
             category="music_player")
    async def _music_ban(self, msg):
        try:
            ban_store = self.storage["banned_users"][str(msg.guild.id)]
        except KeyError:
            ban_store = self.storage["banned_users"][str(msg.guild.id)] = []
        try:
            user = find_user(msg.guild, msg.content.split(None, 1)[1])
        except IndexError:
            user_li = [find_user(msg.guild, x) for x in ban_store]
            banned_list = "\n".join(str(x) if x else "Unknown" for x in user_li)
            if not banned_list:
                banned_list = "None."
            for split_msg in split_message(f"**ANALYSIS: Banned users:**\n```\n{banned_list}```"):
                await respond(msg, split_msg)
            return
        if user:
            if user.id in ban_store:
                ban_store.remove(user.id)
                await respond(msg, f"**AFFIRMATIVE. User {user} has been unbanned from using the music module.**")
                return
            else:
                ban_store.append(user.id)
            await respond(msg, f"**AFFIRMATIVE. User {user} has been banned from using the music module.**")
        else:
            raise CommandSyntaxError(f"No such user {msg.clean_content.split(None, 1)[1]}")

    @Command("MusicConfig",
             doc="Allows the configuration of per-server music bot options.\n"
                 "Valid options: 'max_queue_length', 'max_video_length', 'vote_skip_threshold', "
                 "'print_queue_on_edit', 'idle_disconnect_time'",
             syntax="[(option) (value)]",
             perms="manage_guild",
             category="music_player")
    async def _music_config(self, msg):
        gid = str(msg.guild.id)
        try:
            opt, val = msg.clean_content.split(None, 2)[1:]
        except ValueError:
            current_conf = "\n".join(f"{k}: {v}" for k, v in self.plugin_config[gid].items())
            await respond(msg, f"**ANALYSIS: Current configuration:**```{current_conf}```")
            return
        opt = opt.lower()
        if opt in ("max_queue_length", "max_video_length", "idle_disconnect_time"):
            try:
                val = int(val)
            except ValueError:
                raise CommandSyntaxError(f"{val} is not a valid integer")
        elif opt == "vote_skip_threshold":
            try:
                val = float(val)
                if not 0 < val <= 1:
                    raise ValueError
            except ValueError:
                raise CommandSyntaxError("Value must be a number between 0 and 1.")
        elif opt == "print_queue_on_edit":
            val = is_positive(val)
        else:
            raise CommandSyntaxError(f"Option {opt} does not exist")
        self.plugin_config[gid][opt] = val
        await respond(msg, f"**AFFIRMATIVE. Option `{opt}` edited to `{val}` successfully.**")

    # Utility functions

    async def on_global_tick(self, _, dt):
        # Cache reaper
        save_required = False
        for file, dl_time in self.storage["downloaded_songs"].copy().items():
            if time() - dl_time > self.plugin_config["video_cache_clear_age"]:
                try:
                    remove_file(file)
                    del self.storage["downloaded_songs"][file]
                    self.logger.debug(f"Deleted old video cache file {file}.")
                    save_required = True
                except OSError:
                    continue
        if save_required:
            self.storage.save()
        # Auto-disconnect
        for player in tuple(self.players.values()):
            await player.idle_check(dt)

    async def create_player(self, voice_channel, text_channel):
        with text_channel.typing():
            voice_client = await voice_channel.connect()
            player = GuildPlayer(self, voice_client, text_channel)
            self.players[voice_channel.guild.id] = player
        return player

    def check_user_permission(self, user, player):
        try:
            if user.id in self.storage["banned_users"].get(str(player.voice_client.guild.id), []):
                raise UserPermissionError("You are banned from using the music player.")
            if user not in player.voice_client.channel.members:
                raise UserPermissionError("You are not in the voice channel.")
        except AttributeError:
            raise UserPermissionError("The bot is not currently in a voice channel.")
        return True

    def get_guild_player(self, msg):
        return self.players.get(msg.guild.id, None)


class GuildPlayer:
    def __init__(self, parent, voice_client, channel):
        self.parent = parent
        self.text_channel = channel
        self.voice_client = voice_client
        self.logger = logging.getLogger(f"red_star.plugin.music_player.player_{self.voice_client.guild.id}")
        self.queue = []
        self.already_played = set()
        self.is_playing = False
        self.current_song = {}
        self.song_mode = SongMode.NORMAL
        self.prev_volume = None
        self._volume = self.parent.plugin_config["default_volume"] / 100
        self._song_start_time = None
        self._song_pause_time = None
        self._skip_votes = set()
        self.gid = str(voice_client.guild.id)
        self._alone_time = 0

    async def prepare_playlist(self, urls):
        with self.text_channel.typing():
            # Fetch video info
            with YoutubeDL(self.parent.ydl_options) as ydl:
                # Create a list of videos to be queued
                to_queue = []
                for url in urls:
                    try:
                        pl_slice = re.match(r"([^{`]+)`*(?:{([^}]*)})?", url)
                        vid_info = await get_running_loop().run_in_executor(None, partial(ydl.extract_info,
                                                                                          pl_slice[1], download=False))
                    except YoutubeDLError as e:
                        await self.text_channel.send(f"**WARNING. An error occurred while downloading video <{url}>. "
                                                     f"It will not be queued.\nError details:** `{e}`")
                        continue
                    # If it's a playlist, we need to flatten it into our to_queue list, and deal with the slicing
                    if vid_info.get("_type") == "playlist":
                        # Slice handling
                        if pl_slice[2]:
                            pl_slices = pl_slice[2].split(",")
                            for s in pl_slices:
                                # Ranges handling
                                if "-" in s:
                                    s_start, s_end = s.split("-", 1)
                                    try:
                                        s_start = int(s_start) - 1
                                    except ValueError:
                                        s_start = 0
                                    try:
                                        s_end = int(s_end)
                                        if s_end > len(vid_info["entries"]):
                                            raise ValueError
                                    except ValueError:
                                        s_end = len(vid_info["entries"])
                                    if s_end < s_start:
                                        continue
                                    to_queue.extend(vid_info["entries"][s_start:s_end])
                                # Single values
                                else:
                                    try:
                                        s = int(s) - 1
                                        to_queue.append(vid_info["entries"][s])
                                    except (ValueError, IndexError):
                                        continue
                        # No slice
                        else:
                            to_queue.extend(vid_info["entries"])
                    # If it's just a video, throw it in
                    else:
                        to_queue.append(vid_info)
            # Queuing is handle by _enqueue_playlist function. Don't bother it if we got nothing.
            if len(to_queue) > 0:
                create_task(self._enqueue_playlist(to_queue))
                return

    async def _enqueue_playlist(self, entries):
        time_until_song = ""
        if len(self.queue) > 0:
            time_until_song = self.queue_duration + (self.current_song.get("duration", 0) - self.play_time)
            time_until_song = f"\nTime until your song: {pretty_duration(time_until_song)}"
            await self.text_channel.send(f"**ANALYSIS: Attempting to queue {len(entries)} videos. Your playback will "
                                         f"begin shortly.{time_until_song}**")
        orig_len = len(self.queue)
        with self.text_channel.typing():
            with YoutubeDL(self.parent.ydl_options) as ydl:
                for vid in entries:
                    # We only want to extract info if we don't already have it. Things get a little funky otherwise.
                    if vid.get("_type") in ("url", "url_transparent"):
                        try:
                            vid = await get_running_loop().run_in_executor(None, partial(ydl.extract_info, vid["url"],
                                                                                         download=False))
                        # Skip broken videos while trying the rest
                        except YoutubeDLError as e:
                            await self.text_channel.send(f"**WARNING. An error occurred while downloading video "
                                                         f"<{vid['url']}>. It will not be queued.\nError details:** "
                                                         f"`{e}`")
                            continue
                    # Abort once the queue is full
                    if len(self.queue) >= get_guild_config(self.parent, self.gid, "max_queue_length"):
                        await self.text_channel.send(f"**WARNING: The queue is full. No more videos will be added.**")
                        break
                    # Skip over videos that are too long
                    elif vid.get("duration", 0) > get_guild_config(self.parent, self.gid, "max_video_length"):
                        max_len = pretty_duration(get_guild_config(self.parent, self.gid, "max_video_length"))
                        await self.text_channel.send(f"**WARNING: Video {vid['title']} exceeds the maximum video"
                                                     f" length ({max_len}). It will not be added.**")
                        continue
                    try:
                        await self._process_video(vid)
                    except TypeError:
                        continue
                    if not self.is_playing:
                        try:
                            create_task(self._play())
                        except ClientException:
                            pass
        await self.text_channel.send(f"**ANALYSIS: Queued {len(self.queue) - orig_len} videos.**")
        if get_guild_config(self.parent, self.gid, "print_queue_on_edit") and self.queue:
            final_msg = f"**ANALYSIS: Current queue:**{self.print_queue()}"
            for msg in split_message(final_msg):
                await self.text_channel.send(msg)

    async def _process_video(self, vid):
        if not vid:
            raise TypeError
        if self.parent.plugin_config["save_audio"] and not vid.get("is_live", False):
            with YoutubeDL(self.parent.ydl_options) as ydl:
                try:
                    await get_running_loop().run_in_executor(None, partial(ydl.process_info, vid))
                    vid["filename"] = ydl.prepare_filename(vid)
                    self.parent.storage["downloaded_songs"][vid["filename"]] = time()
                    self.parent.storage.save()
                except YoutubeDLError as e:
                    await self.text_channel.send(f"**WARNING. An error occurred while downloading video "
                                                 f"{vid['title']}. It will not be queued.\nError details:** `{e}`")
                    return
        vid.setdefault("title", "Unknown")
        vid.setdefault("is_live", False)
        vid.setdefault("duration", 0)
        if vid["duration"] < 0:
            vid["duration"] = 0
        self.queue.append(vid)

    async def _play(self):
        try:
            if self.song_mode == SongMode.SHUFFLE_REPEAT:
                try:
                    vids = {i for i, _ in enumerate(self.queue)}
                    eligible_songs = vids - self.already_played
                    if not eligible_songs:
                        self.already_played.clear()
                        eligible_songs = vids
                    eligible_songs = tuple(eligible_songs)
                    next_song = eligible_songs[randint(0, len(eligible_songs) - 1)]
                    self.already_played.add(next_song)
                    next_song = self.queue[next_song]
                except ValueError:
                    next_song = self.queue[0]
            elif self.song_mode == SongMode.SHUFFLE:
                try:
                    next_song = self.queue.pop(randint(0, len(self.queue) - 1))
                except ValueError:
                    next_song = self.queue.pop()
            elif self.song_mode == SongMode.REPEAT_QUEUE:
                if self.current_song:
                    self.queue.append(self.current_song)
                next_song = self.queue.pop(0)
            elif self.song_mode == SongMode.REPEAT_SONG:
                if self.current_song:
                    next_song = self.current_song
                else:
                    next_song = self.queue.pop(0)
            else:
                next_song = self.queue.pop(0)
        except IndexError:
            await self.text_channel.send("**ANALYSIS: Queue complete.**")
            self.is_playing = False
            self.current_song = {}
            self._skip_votes = set()
            self.voice_client.stop()
            try:
                self.voice_client.source.cleanup()
            except AttributeError:
                pass
            return
        before_args = ""
        if self.parent.plugin_config["save_audio"] and not next_song["is_live"]:
            file = next_song["filename"]
        else:
            file = next_song["url"]
            before_args += " -reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 30"
        self.is_playing = True
        if self.prev_volume:
            self.volume = self.prev_volume
            self.prev_volume = None
            await self.text_channel.send(f"**ANALYSIS: Volume reset to {int(self.volume * 100)}%.**")
        source = PCMVolumeTransformer(FFmpegPCMAudio(file, before_options=before_args, options="-vn"),
                                      volume=self._volume)
        self.voice_client.play(source, after=partial(self._after, loop=get_running_loop()))
        self._song_start_time = time()
        self.current_song = next_song
        self._skip_votes = set()
        await self.text_channel.send(f"**NOW PLAYING: {next_song['title']}.**")

    def toggle_pause(self):
        if self.voice_client.is_paused():
            self.voice_client.resume()
            self._song_start_time += (time() - self._song_pause_time)
            return False
        else:
            self.voice_client.pause()
            self._song_pause_time = time()
            return True

    def _after(self, error, loop):
        if error:
            self.logger.error(error)
        run_coroutine_threadsafe(self._play(), loop)

    def stop(self):
        self.is_playing = False
        self._skip_votes = set()
        try:
            self.voice_client.source.cleanup()
        except AttributeError:
            pass
        self.voice_client.stop()

    @property
    def play_time(self):
        try:
            if self.voice_client.is_paused():
                return self._song_pause_time - self._song_start_time
            else:
                return time() - self._song_start_time
        except TypeError:
            return 0

    @property
    def progress(self):
        try:
            fraction = self.play_time / self.current_song["duration"]
        except ZeroDivisionError:
            fraction = 0
        return self.play_time, self.current_song["duration"], fraction

    async def skip_vote(self, uid):
        if uid not in self._skip_votes:
            self._skip_votes.add(uid)
        else:
            await self.text_channel.send("**NEGATIVE. Skip vote already recorded.**")
            return
        total_users = len(self.voice_client.channel.members) - 1  # Don't want to count the bot itself
        threshold = get_guild_config(self.parent, self.gid, "vote_skip_threshold")
        vote_count = len(self._skip_votes)
        if vote_count / total_users >= threshold:
            await self.text_channel.send("**AFFIRMATIVE. Skipping current song.**")
            self.stop()
        else:
            votes_needed = ceil(total_users * threshold) - vote_count
            await self.text_channel.send(f"**AFFIRMATIVE. Skip vote recorded. {votes_needed} votes needed to skip.**")

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, val):
        self._volume = val
        try:
            self.voice_client.source.volume = val
        except AttributeError:
            pass

    @property
    def queue_duration(self):
        try:
            return sum(vid["duration"] for vid in self.queue)
        except TypeError:
            return 0

    def print_queue(self):
        str_list = []
        for i, vid in enumerate(self.queue):
            duration = fixed_width_duration(vid.get("duration", 0))
            # Title truncating and padding
            title = vid.get("title", "Unknown")
            if len(title) >= 59:
                title = title[:56] + "..."
            str_list.append(f"[{i+1:02d}][{title:-<59}][{duration}]")
        return "```\n{}```".format("\n".join(str_list))

    def print_now_playing(self):
        play_time, duration, progress = self.progress
        dur_str = fixed_width_duration(play_time) + "/" + fixed_width_duration(duration)
        progbar = progress_bar(progress, 70)
        title = self.current_song.get("title", "Unknown")
        if len(title) > 57:
            title = title[:54] + "..."
        return f"[{title:-<57}][{dur_str}]\n[{progbar}]"

    async def idle_check(self, dt):
        alone = len(self.voice_client.channel.members) <= 1
        if not alone:
            self._alone_time = 0
            return
        self._alone_time += dt
        if self._alone_time > get_guild_config(self.parent, self.gid, "idle_disconnect_time"):
            self.stop()
            await self.voice_client.disconnect()
            del self.parent.players[int(self.gid)]


def seconds_to_minutes(secs, hours=False):
    mn, sec = divmod(secs, 60)
    if hours:
        hr, mn = divmod(mn, 60)
        return int(hr), int(mn), int(ceil(sec))
    else:
        return int(mn), int(ceil(sec))


def pretty_duration(seconds):
    if seconds <= 0:
        return "Unknown"
    elif seconds >= 3600:
        hours, minutes, seconds = seconds_to_minutes(seconds, hours=True)
        return f"{hours}:{minutes:02d}:{seconds:02d}"
    else:
        minutes, seconds = seconds_to_minutes(seconds)
        return f"{minutes:02d}:{seconds:02d}"


def fixed_width_duration(seconds):
    minutes, seconds = seconds_to_minutes(seconds)
    if seconds <= 0:
        return "--:--"
    if minutes > 99:
        if seconds >= 30:
            minutes += 1
        return f"{minutes:04d}m"
    else:
        return f"{minutes:02d}:{seconds:02d}"


def progress_bar(progress, length=70):
    if not 0 <= progress <= 1:
        raise ValueError("First argument is expected to be a floating-point between 0 and 1.")
    bars = floor(length * progress)
    return "█" * bars + "-" * (length - bars)


class SongMode(enum.Enum):
    NORMAL = "normal"
    REPEAT_SONG = "repeat_song"
    REPEAT_QUEUE = "repeat_queue"
    SHUFFLE = "shuffle"
    SHUFFLE_REPEAT = "shuffle_repeat"

    def __str__(self):
        return self.name



import json
import re
from asyncio import sleep
from discord import Embed
from string import capwords
from urllib.request import urlopen
from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import UserPermissionError
from red_star.rs_utils import respond
from red_star.rs_version import version, VersionInfo, version_tuple
from red_star.command_dispatcher import Command


class Info(BasePlugin):
    name = "info"
    version = "1.1"
    author = "medeor413"
    description = "A plugin that provides commands for fetching information about other commands, or the bot itself."
    default_config = {
        "message_maintainers_when_update_available": True
    }

    async def activate(self):
        self.commands = {}
        self.categories = {}

    async def on_all_plugins_loaded(self):
        await sleep(1)
        await self.build_help()
        await self.check_for_updates()

    async def on_plugin_activated(self, _):
        await sleep(1)
        await self.build_help()

    async def on_plugin_deactivated(self, _):
        await sleep(1)
        await self.build_help()

    async def build_help(self):
        self.commands = self.client.command_dispatcher.commands
        self.categories = {}
        for command in self.commands.values():
            name = command.name
            cmd_category = command.category.lower()
            if cmd_category not in self.categories:
                self.categories[cmd_category] = {}
            self.categories[cmd_category][name] = command

    async def check_for_updates(self):
        repo_url = "https://api.github.com/repos/Medeor413/Red_Star/releases/latest"
        response = urlopen(repo_url)
        if response.status == 200:
            response = json.load(response)
            ver = re.match(r".*(\d+)\.(\d+)\.(\d+)", response["tag_name"])
            latest_version = VersionInfo(major=int(ver[1]), minor=int(ver[2]), patch=int(ver[3]),
                                         releaselevel="release")
            if latest_version > version_tuple:
                self.logger.warning(f"Red Star is out of date!\n"
                                    f"Running version: {version}; latest version: {ver[0]}\n"
                                    f"Please update Red Star as soon as possible.")
                if self.plugin_config["message_maintainers_when_update_available"]:
                    maintainers = [self.client.get_user(i) for i in
                                   self.config_manager.config.get("bot_maintainers", [])]
                    for user in maintainers:
                        await user.send(f"**WARNING: Red Star is out of date!\n"
                                        f"Running version: {version}; latest version: {ver[0]}\n"
                                        f"Please update Red Star as soon as possible.**")

    @Command("Help",
             doc="Displays information on commands.",
             syntax="[category/command]",
             category="info")
    async def _help(self, msg):
        if not self.categories:
            await self.build_help()
        try:
            search = msg.clean_content.split(None, 1)[1].lower()
        except IndexError:
            categories = "\n".join(sorted([capwords(x, "_") for x in self.categories.keys()]))
            await respond(msg, f"**ANALYSIS: Command categories:**```\n{categories}\n```")
            return
        user_perms = {x for x, y in msg.author.guild_permissions if y}
        if search in [x.lower() for x in self.commands.keys()]:
            cmd = self.commands[search]
            name = cmd.name
            syntax = cmd.syntax
            if not syntax:
                syntax = "N/A"
            doc = cmd.__doc__
            perms = cmd.perms
            category = capwords(cmd.category, "_")
            aliases = f"(Aliases: {', '.join(cmd.aliases)})" if cmd.aliases else ""
            if not (user_perms >= perms or self.config_manager.is_maintainer(msg.author)):
                raise UserPermissionError
            text = f"**ANALYSIS: Command {name}:**```\n{name} (Category {category}) {aliases}\n\n{doc}\n\n" \
                   f"Syntax: {syntax}\n```"
            await respond(msg, text)
        elif search in self.categories.keys():
            name = capwords(search, "_")
            cmds = {x.name for x in self.categories[search].values()
                    if user_perms >= x.perms or self.config_manager.is_maintainer(msg.author)}
            cmds = sorted(list(cmds))
            if cmds:
                text = "\n".join(cmds)
                await respond(msg, f"**ANALYSIS: Category {name}:**```\n{text}\n```")
            else:
                await respond(msg, "**WARNING: You do not have permission for any command in this category.**")
        else:
            await respond(msg, f"**WARNING: No such category or command {search}**")

    @Command("About",
             doc="Displays information about the bot.",
             category="info")
    async def _about(self, msg):
        deco = self.client.command_dispatcher.conf[str(msg.guild.id)]["command_prefix"]
        desc = f"Red Star: General purpose command AI for Discord.\n" \
               f"Use {deco}help for command information."
        em = Embed(title="About Red Star", color=0xFF0000, description=desc)
        em.set_thumbnail(url="https://raw.githubusercontent.com/medeor413/Red_Star/master/default_avatar.png")
        em.add_field(name="GitHub", value="https://github.com/medeor413/Red_Star")
        em.add_field(name="Version", value=version)
        await respond(msg, embed=em)



import re
import shlex
from discord import NotFound, Forbidden
from red_star.plugin_manager import BasePlugin
from red_star.rs_utils import respond, find_user, RSArgumentParser
from red_star.rs_errors import CommandSyntaxError
from red_star.command_dispatcher import Command


class AdminCommands(BasePlugin):
    name = "admin_commands"
    version = "1.1.1"
    author = "medeor413"
    description = "A plugin that adds useful administrative commands. Currently only features Purge."
    log_events = {"purge_event"}

    @Command("Purge", "Prune",
             doc="Purges messages from the channel in bulk.\nUse -r option for regexp match filtering.\nWARNING: "
                 "some special characters such as \"\\\" may need to be escaped - eg, use \"\\\\\" or wrap match "
                 "into quotation marks instead.",
             syntax="(count) [match] [-u/--user mention/ID/Name] [-r/--regex] [-v/--verbose] [-e/--emulate/--dryrun]"
                    "[-b/--before message_id] [-a/--after message_id]",
             run_anywhere=True,
             delcall=True,
             perms={"manage_messages"},
             category="admin")
    async def _purge(self, msg):

        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("count", type=int)
        parser.add_argument("match", default=False, nargs='*')
        parser.add_argument("-r", "--regex", action="store_true")
        parser.add_argument("-u", "--user", action="append")
        parser.add_argument("-v", "--verbose", action="store_true")
        parser.add_argument("-e", "--emulate", "--dryrun", action="store_true")
        parser.add_argument("-b", "--before", type=int, default=None)
        parser.add_argument("-a", "--after", type=int, default=None)

        args = parser.parse_args(shlex.split(msg.content))

        if not 0 < args.count <= 250:
            raise CommandSyntaxError("Count must be between 1 and 250.")

        if args['match']:
            args['match'] = ' '.join(args['match'])

        # find all possible members mentioned
        members = []
        if args['user']:
            for q in args['user']:
                u = find_user(msg.guild, q)
                if u:
                    members.append(u.id)

        if args['before']:
            try:
                before_msg = await msg.channel.fetch_message(args['before'])
            except NotFound:
                raise CommandSyntaxError(f"No message in channel with ID {args['before']}.")
        else:
            before_msg = msg

        if args['after']:
            try:
                after_msg = await msg.channel.fetch_message(args['after'])
            except NotFound:
                raise CommandSyntaxError(f"No message in channel with ID {args['before']}.")
        else:
            after_msg = None

        def check(check_msg):
            if members and check_msg.author.id not in members:
                return False
            if args.regex:
                return self.match_regex(check_msg, args.match)
            else:
                return self.match_simple(check_msg, args.match)

        if not args['emulate']:
            # actual purging
            if not msg.channel.permissions_for(msg.guild.me).manage_messages:
                raise Forbidden
            deleted = await msg.channel.purge(limit=args['count'], before=before_msg, after=after_msg, check=check)
        else:
            # dry run to test your query
            deleted = [m async for m in msg.channel.history(limit=args['count'], before=before_msg,
                                                            after=after_msg).filter(check)]

        # if you REALLY want those messages
        if args['verbose']:
            await self.plugin_manager.hook_event("on_log_event", msg.guild,
                                                 "**WARNING: Beginning verbose purge dump.**",
                                                 log_type="purge_event")
            for d in deleted[::-1]:
                await self.plugin_manager.hook_event("on_log_event", msg.guild,
                                                     f"`{d.author}({d.author.id}) @ {d.created_at}:`\n{d.content}",
                                                     log_type="purge_event")
            await self.plugin_manager.hook_event("on_log_event", msg.guild,
                                                 "**Verbose purge dump complete.**",
                                                 log_type="purge_event")

        await respond(msg, f"**PURGE COMPLETE: {len(deleted)} messages purged.**" +
                      (f"\n**Purge query: **{args['match']}" if args['match'] else ""), delete_after=5)

    @staticmethod
    def match_simple(msg, searchstr):
        return not searchstr or searchstr.lower() in msg.content.lower()

    @staticmethod
    def match_regex(msg, searchstr):
        return not searchstr or re.match(searchstr.lower(), msg.content.lower())



from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import CommandSyntaxError, UserPermissionError
from red_star.rs_utils import respond, RSArgumentParser
from red_star.command_dispatcher import Command
import shlex
from discord import Embed, Message


class Voting(BasePlugin):
    name = "voting"
    version = "1.0"
    author = "GTG3000"
    description = "A plugin for creating, voting in, and automatically tallying the results of polls."

    polls = {}  # dict of lists of polls {gid:{msg.id:Poll}}

    class Poll:

        _abc = "abcdefghijklmnopqrst"
        _emo = "🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇳🇴🇵🇶🇷🇸🇹"
        _a_e = dict(zip(_abc, _emo))  # easy letter to emote conversion
        _e_a = dict(zip(_emo, _abc))  # easy emote to letter conversion

        message = None  # message that contains the voting options
        author = None  # author ID (we don't really need anything else)
        id = 0  # message.id, for speed
        hid = ""  # human-readable ID
        votes = None  # dict of lists of options voted for - member.id : ["a", "b", "c"]
        vote_count = None  # dict of vote numbers
        vote_limit = 0  # 0 for no limit
        query = ""  # the question
        options = None  # dict of strings to vote for {"a":"thing", "b":"other thing"} limit 20 (by the reaction limit)
        active = False
        allow_retracting = True

        def __init__(self, msg, hid=None, vote_limit=1, author=None, allow_retracting=True):
            """
            :type msg:discord.Message
            :param msg:
            :param hid:
            """
            self.message = msg
            self.author = author
            self.id = msg.id
            self.hid = hid if hid else str(msg.id)
            self.votes = {}
            self.vote_count = {}
            self.vote_limit = vote_limit
            self.allow_retracting = allow_retracting
            self.options = {}

        def setquery(self, query):
            """
            :type query:str
            :param query: string of the voting question/query
            """
            self.query = query

        async def add_option(self, option):
            if len(self.options) < 20:
                self.vote_count[self._abc[len(self.options)]] = 0
                await self.message.add_reaction(self._emo[len(self.options)])
                self.options[self._abc[len(self.options)]] = option

        async def update(self):
            await self.message.edit(content="", embed=self._build_embed())

        def _build_embed(self):
            t_embed = Embed(type="rich", colour=16711680)
            t_embed.title = f"\"{self.hid}\""
            t_embed.description = f"{self.query}\n\nBy <@{self.author}>"
            for k, v in self.options.items():
                t_embed.add_field(name=f"{self._a_e[k]}", value=f"{v} : {self.vote_count[k]}")
            return t_embed

        async def add_reaction(self, reaction, user):
            if isinstance(reaction.emoji, str) and self.active:
                if reaction.emoji not in self._emo or not await self.vote(self._e_a[reaction.emoji], user):
                    await reaction.message.remove_reaction(reaction.emoji, user)

        async def remove_reaction(self, reaction, user):
            if isinstance(reaction.emoji, str) and reaction.emoji in self._emo and self.active:
                await self.vote(self._e_a[reaction.emoji], user, False)

        async def vote(self, option, user, up=True):
            if user.id not in self.votes:
                self.votes[user.id] = set()
            if up:
                if option not in self.votes[user.id]:
                    if len(self.votes[user.id]) < self.vote_limit or self.vote_limit == 0:
                        self.votes[user.id].add(option)
                        self.vote_count[option] += 1
                        await self.update()
                        return True
                    else:
                        return False
            elif option in self.votes[user.id] and self.allow_retracting:
                self.votes[user.id].remove(option)
                self.vote_count[option] -= 1
                await self.update()
                return True
            else:
                return False

    @Command("StartVote",
             syntax="(HID) (Query) [Questions, up to 20] [-1/--question, up to 20] [-v/vote_limit, integer, "
                    "0 for no limit] [-n/--no_retracting, to disallow removing votes]",
             doc="Starts a new vote, with provided Human-readableID, Query and Questions.\n"
                 "Use -v to allow users to vote for more than one option and -n to prevent users from changing their"
                 "mind.\nUses shlex for splitting, so multiple words can be wrapped into \"\".\n"
                 "HID is used to interact with the poll through other commands, keep it one word.",
             run_anywhere=True,
             category="voting")
    async def _startvote(self, msg: Message):
        """
        Generates a vote, posts a vote embed.
        :type msg: discord.Message
        :param msg:
        :return:
        """
        args = shlex.split(msg.clean_content)
        gid = str(msg.guild.id)

        if gid not in self.polls:
            self.polls[gid] = {}

        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("poll_hid")
        parser.add_argument("query")
        parser.add_argument("questions", default=[], nargs='*')
        parser.add_argument("--question", "-q", default=[], action="append")
        parser.add_argument("--vote_limit", "-v", default=1, type=int)
        parser.add_argument("--no_retracting", "-n", action="store_false")

        args = parser.parse_args(args)

        t_poll = self.Poll(await respond(msg, "**Generating poll message.**"),
                           hid=args['poll_hid'].lower(),
                           author=msg.author.id,
                           vote_limit=args['vote_limit'],
                           allow_retracting=args['no_retracting'])
        t_poll.setquery(args['query'])
        for opt in [*args['questions'], *args['question']]:
            await t_poll.add_option(opt)

        await t_poll.update()
        t_poll.active = True
        self.polls[gid][t_poll.message.id] = t_poll

    @Command("EndVote", run_anywhere=True,
             syntax="(HID)",
             doc="Ends the given vote. You must be the creator of the vote to end it.\n"
                 "Alternatively, you must have manage_messages permission or be a bot maintainer.",
             category="voting")
    async def _endvote(self, msg: Message):
        args = msg.clean_content.split(maxsplit=1)
        gid = str(msg.guild.id)

        if gid not in self.polls:
            await respond(msg, "**WARNING: No polls active in current server.**")
            return
        elif len(args) > 1:
            candidates = [(k, c) for k, c in self.polls[gid].items()
                          if c.hid == args[1].lower() or str(c.id) == args[1]]
            results = []
            for k, c in candidates:
                if c.author != msg.author.id and \
                        msg.author.id not in self.config_manager.config.get("bot_maintainers", []) and \
                        not msg.channel.permissions_for(msg.author).manage_messages:
                    continue
                max_votes = sorted(c.vote_count.items(), key=lambda x: x[1]).pop()[1]
                winners = '\n'.join(c.options[k] for k, v in c.vote_count.items() if v == max_votes)
                results.append(f'Query: {c.query}. With {max_votes} votes, leading results:\n{winners}')
                del self.polls[gid][k]
            if not results:
                raise UserPermissionError
            results = '\n\n'.join(results)
            await respond(msg, f"**AFFIRMATIVE. ANALYSIS: {len(candidates)} polls terminated with results:**"
                               f"```{results}```")
        else:
            raise CommandSyntaxError("WARNING: Poll HID or ID required")

    @Command("Vote", "UpVote", "DownVote", run_anywhere=True, delcall=True,
             syntax="(hid) (option, single letter from a to t)",
             doc="Allows users to vote without using reactions.\n"
                 "Use \"DownVote\" variant to remove your vote, if possible.",
             category="voting")
    async def _vote(self, msg: Message):
        args = shlex.split(msg.clean_content)
        gid = str(msg.guild.id)

        if gid in self.polls:
            candidates = [k for k, c in self.polls[gid].items() if c.hid == args[1].lower()]
            if len(candidates) > 1:
                raise CommandSyntaxError("**WARNING: More than one poll with same HID**")
            else:
                up = not args[0].lower().endswith('downvote')
                p = self.polls[gid][candidates[0]]
                if len(args[2]) == 1 and args[2].lower() in "abcdefghijklmnopqrst":
                    result = await p.vote(args[2].lower(), msg.author, up)
                    if not result:
                        if up:
                            await respond(msg, "**NEGATIVE: Out of votes.**", delete_after=5)
                        else:
                            await respond(msg, "**NEGATIVE: Could not remove vote.**", delete_after=5)
                else:
                    raise CommandSyntaxError(f"Incorrect voting option {args[2]}")

    async def on_reaction_add(self, reaction, user):
        """
        :type reaction:discord.Reaction
        :param reaction:
        :param user:
        :return:
        """
        gid = str(reaction.message.guild.id)
        if gid in self.polls and reaction.message.id in self.polls[gid]:
            await self.polls[gid][reaction.message.id].add_reaction(reaction, user)

    async def on_reaction_remove(self, reaction, user):
        """
        :type reaction:discord.Reaction
        :param reaction:
        :param user:
        :return:
        """
        gid = str(reaction.message.guild.id)
        if gid in self.polls and reaction.message.id in self.polls[gid]:
            await self.polls[gid][reaction.message.id].remove_reaction(reaction, user)



from collections import namedtuple
__all__ = ['version_tuple', 'version', 'VersionInfo']


class VersionInfo(namedtuple("VersionInfo", 'major minor patch releaselevel')):
    def as_string(self):
        verstr = f"{self.major}.{self.minor}.{self.patch}"
        if self.releaselevel != "release":
            verstr += f"-{version_tuple.releaselevel}"
        return verstr


version_tuple = VersionInfo(major=2, minor=2, patch=3, releaselevel="release")

version = version_tuple.as_string()



from red_star.rs_errors import ChannelNotFoundError


class ChannelManager:
    def __init__(self, client):
        self.client = client
        self.config_manager = client.config_manager
        self.conf = self.config_manager.get_plugin_config_file("channel_manager.json")
        self.channel_types = {"commands"}
        self.channel_categories = {"no_read"}
        if "channel_manager" in self.config_manager.config:  # Port from the old config.json storage
            self.conf.update(self.config_manager.config["channel_manager"])
            del self.config_manager.config["channel_manager"]
            self.config_manager.save_config()
        self.default_config = {
            "channels": {},
            "categories": {}
        }

    def add_guild(self, gid):
        if gid not in self.conf:
            self.client.logger.info(f"Registered new guild: {gid}")
            self.conf[gid] = self.default_config.copy()
        guild_conf = self.conf[gid]
        new_channels = {i: None for i in self.channel_types}
        new_channels.update(guild_conf["channels"])
        guild_conf["channels"] = new_channels
        new_categories = {i: [] for i in self.channel_categories}
        new_categories.update(guild_conf["categories"])
        guild_conf["categories"] = new_categories
        self.conf.save()

    def get_channel(self, guild, chantype):
        gid = str(guild.id)
        chantype = chantype.lower()
        chan = self.conf[gid]["channels"][chantype]
        chan = self.client.get_channel(chan)
        if not chan:
            raise ChannelNotFoundError(chantype)
        return chan

    def set_channel(self, guild, chantype, channel):
        gid = str(guild.id)
        chantype = chantype.lower()
        if channel:
            self.conf[gid]["channels"][chantype] = channel.id
        else:
            self.conf[gid]["channels"][chantype] = None
        self.conf.save()

    def get_category(self, guild, category: str):
        guild_categories = self.conf[str(guild.id)]["categories"]
        if category.lower() in guild_categories:
            return guild_categories[category.lower()]
        else:
            return None

    def channel_in_category(self, guild, category, channel):
        guild_categories = self.conf[str(guild.id)]["categories"]
        if category.lower() not in guild_categories:
            return False
        if channel.id not in guild_categories[category.lower()]:
            return False
        return True

    def add_channel_to_category(self, guild, category, channel):
        category = self.conf[str(guild.id)]["categories"].setdefault(category.lower(), [])
        if channel.id not in category:
            category.append(channel.id)
            self.conf.save()
            return True
        else:
            return False

    def remove_channel_from_category(self, guild, category, channel):
        gid = str(guild.id)
        category = category.lower()
        if self.channel_in_category(guild, category, channel):
            self.conf[gid]["categories"][category].remove(channel.id)
            self.conf.save()
            return True
        else:
            return False



from red_star.plugin_manager import BasePlugin
from discord import utils, VoiceChannel
from red_star.rs_utils import respond
from red_star.command_dispatcher import Command
from red_star.rs_errors import ChannelNotFoundError, CommandSyntaxError
import shlex


class ChannelManagerCommands(BasePlugin):
    name = "channel_manager_commands"
    version = "1.0.3"
    author = "medeor413"
    description = "A plugin that provides commands for interfacing with Red Star's channel_manager."
    default_config = {}

    @Command("GetChannel",
             doc="Gets information on the specified channel type (or all channel types if none specified) in this "
                 "server.",
             syntax="[channel type]",
             category="channel_management",
             perms={"manage_guild"})
    async def _get_channel_cmd(self, msg):
        gid = str(msg.guild.id)
        try:
            chantype = msg.clean_content.split(None, 1)[1].lower()
            try:
                chan = self.channel_manager.get_channel(msg.guild, chantype)
                await respond(msg, f"**ANALYSIS: The {chantype} channel for this server is {chan.mention}.**")
            except ChannelNotFoundError:
                await respond(msg, f"**ANALYSIS: No channel of type {chantype} set for this server.**")
        except IndexError:
            chantypes = "\n".join([f"{x.capitalize()}: {self.client.get_channel(y).name if y else 'Unset'}"
                                   for x, y in self.channel_manager.conf[gid]['channels'].items()])
            await respond(msg, f"**ANALYSIS: Channel types for this server:**```\n{chantypes}```")

    @Command("SetChannel",
             doc="Sets the specified channel type to the specified channel for this server or disables it.\n"
                 "Use channel mention for text channels or channel name for voice channels.\n"
                 "Voice channel types must be prefixed by \"Voice\".",
             syntax="(chantype) [channel]",
             category="channel_management",
             perms={"manage_guild"},
             run_anywhere=True)
    async def _set_channel_cmd(self, msg):
        args = shlex.split(msg.content)

        if len(args) > 1:
            chantype = args[1].lower()
        else:
            raise CommandSyntaxError("No channel type provided.")

        if chantype not in self.channel_manager.channel_types:
            type_list = "\n".join(x.capitalize() for x in self.channel_manager.channel_types)
            await respond(msg, f"**WARNING: No such channel type {chantype}. Channel types:**\n"
                               f"```\n{type_list}\n```")
            return

        if len(args) > 2:
            if chantype.startswith("voice"):
                channel = args[2].lower()
                channel = utils.find(lambda x: isinstance(x, VoiceChannel) and x.name.lower() == channel,
                                     msg.guild.channels)
                if not channel:
                    raise CommandSyntaxError(f"Voice channel {args[2].lower()} not found.")
            else:
                if msg.channel_mentions:
                    channel = msg.channel_mentions[0]
                else:
                    raise CommandSyntaxError("No channel provided.")
        else:
            channel = None

        self.channel_manager.set_channel(msg.guild, chantype, channel)

        if channel:
            await respond(msg, f"**ANALYSIS: The {chantype} channel for this server has been set to "
                               f"{channel.mention}.**")
        else:
            await respond(msg, f"**ANALYSIS: The {chantype} channel for this server has been disabled.**")

    @Command("GetCategory",
             doc="Gets the members of the specified channel category on this server.",
             syntax="[category]",
             category="channel_management",
             perms={"manage_guild"})
    async def _get_category(self, msg):
        gid = str(msg.guild.id)
        try:
            category = msg.clean_content.split(None, 1)[1].lower()
            if category in self.channel_manager.conf[gid]['categories']:
                catestr = ", ".join([msg.guild.get_channel(x).name for
                                     x in self.channel_manager.conf[gid]['categories'][category]])
                await respond(msg, f"**ANALYSIS: Category {category} contains the following channels:**\n"
                                   f"```\n{catestr}```")
            else:
                await respond(msg, f"**ANALYSIS: No such category {category}.**")
        except IndexError:
            catestr = "\n".join(self.channel_manager.conf[gid]['categories'].keys())
            await respond(msg, f"**ANALYSIS: Available categories:**\n```\n{catestr}```")

    @Command("AddToCategory",
             doc="Adds the given channel to the specified category for this server.\n"
                 "Use channel mention for text channels or channel name for voice channels.\n"
                 "Voice channel categories must be prefixed by \"voice\".\n"
                 "Use --create to force addition of a new category.",
             syntax="[-c/--create](category) (channel)",
             category="channel_management",
             perms={"manage_guild"})
    async def _add_to_category(self, msg):
        args = shlex.split(msg.content)
        ignore_missing = False
        try:
            category = args[1].lower()
            if category in ("-c", "--create"):
                ignore_missing = True
                category = args[2].lower()
        except IndexError:
            raise CommandSyntaxError("No category provided.")

        if category not in self.channel_manager.conf[str(msg.guild.id)]["categories"] and not ignore_missing:
            raise CommandSyntaxError(f"No such channel category {category}.")

        if len(args) > 2:
            res = ""
            if category.startswith("voice"):
                for arg in args[2:]:
                    channel = arg.lower()
                    channel = utils.find(lambda x: x.name.lower() == channel, msg.guild.voice_channels)
                    if not channel:
                        raise CommandSyntaxError(f"Voice channel {args[2].lower()} not found.")
                    if self.channel_manager.add_channel_to_category(msg.guild, category, channel):
                        res = f"{res}{str(channel)}\n"
            else:
                if msg.channel_mentions:
                    for channel in msg.channel_mentions:
                        if self.channel_manager.add_channel_to_category(msg.guild, category, channel):
                            res = f"{res}{str(channel)}\n"
                else:
                    raise CommandSyntaxError("No channel provided")

            await respond(msg, f"**ANALYSIS: The following channels were added to category {category}:**```\n{res}```")
        else:
            raise CommandSyntaxError("No channel provided")

    @Command("RMFromCategory",
             doc="Removes the given channel from the specified category from the server.\n"
                 "Use channel mention for text channels or channel name for voice channels.\n"
                 "Voice channel categories must be prefixed by \"voice\".",
             syntax="(category) (channel)",
             category="channel_management",
             perms={"manage_guild"})
    async def _rm_from_category(self, msg):
        args = shlex.split(msg.content)
        if len(args) > 1:
            category = args[1].lower()
        else:
            raise CommandSyntaxError("No category provided.")

        if category not in self.channel_manager.conf[str(msg.guild.id)]["categories"]:
            raise CommandSyntaxError(f"No such channel category {category}.")

        if len(args) > 2:
            res = ""
            if category.startswith("voice"):
                for arg in args[2:]:
                    channel = arg.lower()
                    channel = utils.find(lambda x: x.name.lower() == channel, msg.guild.voice_channels)
                    if not channel:
                        raise CommandSyntaxError(f"Voice channel {args[2].lower()} not found")
                    if self.channel_manager.remove_channel_from_category(msg.guild, category, channel):
                        res += f"✓ - {str(channel)}\n"
                    else:
                        res += f"✗ - {str(channel)}\n"
            else:
                if msg.channel_mentions:
                    for channel in msg.channel_mentions:
                        if self.channel_manager.remove_channel_from_category(msg.guild, category, channel):
                            res += f"✓ - {str(channel)}\n"
                        else:
                            res += f"✗ - {str(channel)}\n"
                else:
                    raise CommandSyntaxError("No channel provided")

            await respond(msg, f"**ANALYSIS: Processed channel removals from category {category}:**\n```\n{res}```")
        else:
            raise CommandSyntaxError("No channel provided")



from red_star.plugin_manager import BasePlugin
from red_star.rs_errors import ChannelNotFoundError
from red_star.rs_utils import sub_user_data, respond, get_guild_config
from random import choice


class Announcer(BasePlugin):
    name = "announcer"
    version = "1.1"
    author = "medeor413"
    description = "A plugin for basic automated announcements, triggering messages when the bot starts, " \
                  "a new member joins, or the bot is pinged."
    default_config = {
        "default": {
            "greeting_message": "Sample message to be broadcasted on bot load.",
            "new_member_announce_message": "Message for new players. Replaces <username> with their name and "
                                           "<usermention> with a mention.",
            "ping_messages": True,
            "ping_messages_on_everyone": False,
            "ping_message_options": [
                "**STOP**",
                "***STOP THIS AT ONCE!***",
                "**CEASE**",
                "***CEASE THIS AT ONCE!***",
                "**DO NOT**",
                "**HOW** ***DARE*** **YOU!**",
                "**`soft crying`**",
                "**DO NOT PING THE ROBOT!**",
                "**STOP PINGING ME!**",
                "**`INTERNAL SCREAMING`**",
                "**`EXTERNAL SCREAMING`**",
                "**`INTERNAL AND EXTERNAL SCREAMING`**"
            ]
        }
    }
    channel_types = {"startup", "welcome"}

    async def on_all_plugins_loaded(self):
        await self._greet()

    async def _greet(self):
        for guild in self.client.guilds:
            gid = str(guild.id)
            msg = get_guild_config(self, gid, "greeting_message")
            try:
                greet_channel = self.channel_manager.get_channel(guild, "startup")
                await greet_channel.send(msg)
            except ChannelNotFoundError:
                continue

    async def _ping_response(self, msg):
        gid = str(msg.guild.id)
        response = sub_user_data(msg.author, choice(get_guild_config(self, gid, "ping_message_options")))
        await respond(msg, response)

    # Event hooks

    async def on_message(self, msg):
        gid = str(msg.guild.id)
        ping_messages = get_guild_config(self, gid, "ping_messages")
        ping_messages_everyone = get_guild_config(self, gid, "ping_messages_on_everyone")
        if ping_messages and (ping_messages_everyone >= msg.mention_everyone) and msg.guild.me.mentioned_in(msg):
            await self._ping_response(msg)

    async def on_member_join(self, msg):
        gid = str(msg.guild.id)
        text = sub_user_data(msg, get_guild_config(self, gid, "new_member_announce_message"))
        try:
            chan = self.channel_manager.get_channel(msg.guild, "welcome")
            await chan.send(text)
        except ChannelNotFoundError:
            pass



import inspect
import logging
import importlib
import importlib.util
from sys import exc_info, modules
from types import ModuleType


class PluginManager:
    """
    Manages the loading of plugins and dispatching of event hooks.
    """

    def __init__(self, client):
        self.client = client
        self.config_manager = client.config_manager
        self.channel_manager = client.channel_manager
        self.command_dispatcher = client.command_dispatcher
        self.modules = {}
        self.plugins = {}
        self.active_plugins = {}
        self.logger = logging.getLogger("red_star.plugin_manager")
        self.logger.debug("Initialized plugin manager.")
        self.last_error = None
        self.plugin_package = ModuleType("red_star_plugins")
        self.plugin_package.__path__ = []
        modules["red_star_plugins"] = self.plugin_package

    def __repr__(self):
        return f"<PluginManager: Plugins: {self.plugins.keys()}, Active: {self.active_plugins}>"

    def load_all_plugins(self, plugin_paths):
        self.logger.debug("Loading plugins...")
        self.plugin_package.__path__.extend(str(x) for x in plugin_paths)
        for path in plugin_paths:
            self._load_plugin_folder(path)
        self.config_manager.save_config()
        self.logger.info(f"Loaded {len(self.plugins)} plugins from {len(self.modules)} modules.")

    def _load_plugin_folder(self, plugin_path):
        self.logger.debug(f"Loading plugins from {plugin_path}...")
        loaded = set()
        plugin_path.mkdir(parents=True, exist_ok=True)
        for file in plugin_path.iterdir():
            if file.stem.startswith(("_", ".")):
                continue
            if (file.suffix == ".py" or file.is_dir()) and file not in loaded:
                try:
                    modul = self._load_module(file.stem)
                    self.load_plugin(modul)
                    loaded.add(file)
                except (SyntaxError, ImportError):
                    self.logger.exception(f"Exception encountered loading plugin {file.stem}: ", exc_info=True)
                    continue
                except FileNotFoundError:
                    self.logger.error(f"File {file.stem} missing when load attempted!")
                    continue

    def _load_module(self, module_name):
        mod = importlib.import_module(f"red_star_plugins.{module_name}")
        self.modules[module_name] = mod
        self.logger.debug(f"Imported module {module_name}.")
        return mod

    def _get_plugin_class(self, plugin_module):
        def predicate(cls):
            return inspect.isclass(cls) and issubclass(cls, BasePlugin) and cls is not BasePlugin

        class_list = set()
        for name, obj in inspect.getmembers(plugin_module, predicate=predicate):
            obj.client = self.client
            obj.config_manager = self.config_manager
            if obj.default_config:
                self.config_manager.init_plugin_config(obj.name, obj.default_config)
                obj.plugin_config = self.config_manager.get_plugin_config(obj.name)
            obj.channel_manager = self.channel_manager
            if obj.channel_types:
                self.channel_manager.channel_types |= obj.channel_types
            if obj.channel_categories:
                self.channel_manager.channel_categories |= obj.channel_categories
            obj.plugin_manager = self
            obj.plugins = self.active_plugins
            obj.logger = logging.getLogger("red_star.plugin." + obj.name)
            class_list.add(obj)
        return class_list

    def load_plugin(self, plugin_module):
        classes = self._get_plugin_class(plugin_module)
        for cls in classes:
            self.plugins[cls.name] = cls()
            self.logger.debug(f"Loaded plugin {cls.name}")

    async def activate_all(self):
        self.logger.info("Activating plugins.")
        if "disabled_plugins" not in self.config_manager.config:
            self.config_manager.config["disabled_plugins"] = []
            self.config_manager.save_config()
        disabled_plugins = self.config_manager.config["disabled_plugins"]
        for name, plugin in self.plugins.items():
            if name not in self.active_plugins and name not in disabled_plugins:
                self.logger.info("Activating " + plugin.name)
                # noinspection PyBroadException
                try:
                    await plugin.activate()
                    self.active_plugins[name] = plugin
                    self.command_dispatcher.register_plugin(plugin)
                except Exception:
                    self.logger.exception(f"Error occurred while activating plugin {plugin.name}: ", exc_info=True)
        await self.hook_event("on_all_plugins_loaded")

    async def deactivate_all(self):
        self.logger.info("Deactivating plugins.")
        for n, plugin in self.plugins.items():
            if n in self.active_plugins:
                self.logger.info("Deactivating " + plugin.name)
                # noinspection PyBroadException
                try:
                    await plugin.deactivate()
                except Exception:
                    self.logger.exception(f"Error occurred while deactivating plugin {plugin.name}: ", exc_info=True)
                del self.active_plugins[n]
                self.command_dispatcher.deregister_plugin(plugin)

    async def activate(self, name):
        try:
            plg = self.plugins[name]
            if name not in self.active_plugins:
                self.logger.info(f"Activating plugin {name}.")
                # noinspection PyBroadException
                try:
                    await plg.activate()
                    self.active_plugins[name] = plg
                    self.command_dispatcher.register_plugin(plg)
                    await self.hook_event("on_plugin_activated", name)
                except Exception:
                    self.logger.exception(f"Error occurred while activating plugin {name}: ", exc_info=True)
            else:
                self.logger.warning(f"Attempted to activate already active plugin {name}.")
        except KeyError:
            self.logger.error(f"Attempted to activate non-existent plugin {name}.")

    async def deactivate(self, name):
        try:
            plg = self.plugins[name]
            if name in self.active_plugins:
                self.logger.info(f"Deactivating plugin {name}.")
                # noinspection PyBroadException
                try:
                    await plg.deactivate()
                except Exception:
                    self.logger.exception(f"Error occurred while deactivating plugin {name}: ", exc_info=True)
                del self.active_plugins[name]
                self.command_dispatcher.deregister_plugin(plg)
                await self.hook_event("on_plugin_deactivated", name)
            else:
                self.logger.warning(f"Attempted to deactivate already inactive plugin {name}.")
        except KeyError:
            self.logger.error(f"Attempted to deactivate non-existent plugin {name}.")

    async def reload_plugin(self, name):
        try:
            self.logger.info(f"Reloading plugin module {name}.")
            was_active = False
            if name in self.active_plugins:
                was_active = True
                await self.deactivate(name)
            del self.plugins[name]
            modul = self.modules[name]
            importlib.reload(modul)
            self.load_plugin(modul)
            if was_active:
                await self.activate(name)
        except KeyError:
            self.logger.error(f"Attempted to reload non-existent plugin module {name}.")

    async def hook_event(self, event, *args, **kwargs):
        """
        Dispatches an event, with its data, to all plugins.
        :param event: The name of the event. Should match the calling function.
        :param args: Everything that gets passed to the calling function
        should be passed through to this function.
        """
        plugins = set(self.active_plugins.values())
        for plugin in plugins:
            hook = getattr(plugin, event, False)
            if hook:
                # noinspection PyBroadException
                try:
                    await hook(*args, **kwargs)
                except Exception:
                    self.last_error = exc_info()
                    self.logger.exception(f"Exception encountered in plugin {plugin.name} on event {event}: ",
                                          exc_info=True)


class BasePlugin:
    """
    The base plugin class from which all plugin classes must inherit in order to be detected by the plugin manager.
    It is recommended you set all meta-fields (name, description, version, author), but only name is necessary.
    In most circumstances you're going to want the plugin name and the module name to be the same.
    The plugin manager will install various useful things to the class.
    """
    # Metadata fields
    name: str = "Base Plugin"
    description: str = "This is a template class for plugins. Name *must* be filled, other meta-fields are optional."
    version: str = "1.0"
    author: str = "Unknown"
    # Attributes added by plugin manager
    plugins: dict = {}
    plugin_config: dict = {}
    client: "client.RedStar"
    config_manager: "config_manager.ConfigManager"
    channel_manager: "channel_manager.ChannelManager"
    plugin_manager: PluginManager
    logger = logging.Logger
    # User-defined attributes for use internally
    default_config: dict = {}
    channel_types: set = set()
    channel_categories: set = set()

    async def activate(self):
        """
        The method called when the plugin is initialized. Should be used to get all of the Discord-related
        initialization out of the way.
        Raise an exception in this method to cancel activation, say if a required package isn't installed.
        """

    async def deactivate(self):
        """
        The method called when the plugin is uninitalized. Should be used to perform any necessary cleanup.
        """

    def __str__(self):
        """
        Method to return something a little less nasty.
        :return: string: The string to return when str() is called on this object.
        """
        return f"<Plugin {self.name} (Version {self.version})>"

    def __repr__(self):
        """
        Method to return something a little less nasty.
        :return: string: The string to return when repr() is called on this object.
        """
        return f"<Plugin {self.name} (Version {self.version})>"



# Miscellaneous utility functions and classes found here.
import argparse
import re
import json
from red_star.rs_errors import CommandSyntaxError
from urllib.parse import urlparse


class JsonFileDict(dict):
    """
    Dictionary subclass that handles saving the file on edits automatically.
    Try not to instantiate this class directly; instead, use the config_manager's factory method,
    ConfigManager.get_plugin_config_file.
    :param pathlib.Path path: The path that should be saved to.
    """

    def __init__(self, path, json_save_args=None, json_load_args=None, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.json_save_args = {} if json_save_args is None else json_save_args
        self.json_load_args = {} if json_load_args is None else json_load_args
        self.reload()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.save()

    def __delitem__(self, key):
        super().__delitem__(key)
        self.save()

    def save(self):
        with self.path.open("w", encoding="utf-8") as fd:
            json.dump(self, fd, **self.json_save_args)

    def reload(self):
        with self.path.open(encoding="utf-8") as fd:
            self.update(json.load(fd, **self.json_load_args))


class RSNamespace(argparse.Namespace):
    def __getitem__(self, key):
        try:
            return self.__getattribute__(key)
        except AttributeError:
            raise KeyError

    def __setitem__(self, key, value):
        self.__setattr__(key, value)


class RSArgumentParser(argparse.ArgumentParser):

    def __init__(self, add_help=False, ignore_unrecognized_arguments=True, **kwargs):
        self.ignore_unrecognized_arguments = ignore_unrecognized_arguments
        super().__init__(add_help=add_help, **kwargs)

    def exit(self, status=0, message=None):
        raise CommandSyntaxError(message)

    def error(self, message):
        raise CommandSyntaxError(message)

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        if argv and not self.ignore_unrecognized_arguments:
            self.error(f"Unrecognized arguments: {' '.join(argv)}")
        return args

    def parse_known_args(self, args=None, namespace=None):
        if namespace is None:
            namespace = RSNamespace()
        return super().parse_known_args(args=args, namespace=namespace)


def get_guild_config(cls, gid, key):
    """
    Gets guild-specific configuration for an option, or fills it in with the default if unspecified.
    :param BasePlugin cls: The class calling the function, so it can access plugin-specific configs.
    :param str gid: The guild ID of the guild you're working with, as a str.
    :param str key: The config option you're trying to fetch.
    :return: The config option asked for.
    """
    if gid not in cls.plugin_config:
        cls.plugin_config[gid] = cls.plugin_config["default"].copy()
        cls.config_manager.save_config()
    elif key not in cls.plugin_config[gid]:
        cls.plugin_config[gid][key] = cls.plugin_config["default"][key]
        cls.config_manager.save_config()
    return cls.plugin_config[gid][key]


def sub_user_data(user, text):
    """
    Replaces certain tags in data with user info.
    :param user: The User object to get data from.
    :param text: The text string to substitute on.
    :return str: The substituted text.
    """
    rep = {
        "<username>": user.name,
        "<usernick>": user.display_name,
        "<userid>": user.id,
        "<userdiscrim>": user.discriminator,
        "<usermention>": user.mention
    }
    rep = {re.escape(k): v for k, v in rep.items()}
    pattern = re.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[re.escape(m.group(0))], text)
    return text


def find_user(guild, search, return_all=False):
    """
    Convenience function to find users via several checks.
    :param guild: The discord.Guild object in which to search.
    :param search: The search string.
    :param return_all: Whether to return all users that match the criteria or just the first one.
    :return: discord.Member: The Member that matches the criteria, or none.
    """
    funcs = (lambda x: str(x.id) == search, lambda x: x.mention == search, lambda x: str(x).lower() == search.lower(),
             lambda x: x.display_name.lower() == search.lower(), lambda x: x.name.lower() == search.lower())
    final = []
    for func in funcs:
        found = tuple(filter(func, guild.members))
        if found:
            if return_all:
                final += found
            else:
                return found[0]
    if return_all:
        return final


def find_role(guild, search, return_all=False):
    """
    Convenience function to find users via several checks.
    :param guild: The discord.Guild object in which to search.
    :param search: The search string.
    :param return_all: Whether to return all roles that match the criteria or just the first one.
    :return: discord.Role: The Role that matches the criteria, or none.
    """
    funcs = (lambda x: str(x.id) == search, lambda x: x.mention == search,
             lambda x: str(x).lower() == search.lower())
    final = []
    for func in funcs:
        found = tuple(filter(func, guild.roles))
        if found:
            if return_all:
                final += found
            else:
                return found[0]
    if return_all:
        return final


async def respond(msg, response=None, allow_mention_everyone=False, **kwargs):
    """
    Convenience function to respond to a given message. Replaces certain
    patterns with data from the message.
    :param msg: The message to respond to.
    :param response: The text to respond with.
    :param allow_mention_everyone: If True, disables the automatic @everyone and @here filtering. Defaults to False.
    :return discord.Message: The Message sent.
    """
    text = None
    if response:
        if not allow_mention_everyone:  # Filter these out just in case we miss it somehow
            text = response.replace("@everyone", "@\u200beveryone").replace("@here", "@\u200bhere")
        else:
            text = response
        # should've split it first
        # this is just a last-ditch error check
        text = text[:2000]
    elif not kwargs:
        # It's empty, raise an error.
        raise SyntaxError("respond() called without any response arguments.")
    return await msg.channel.send(text, **kwargs)


def split_message(input_string: str, max_len: int = 2000, splitter: str = "\n"):
    """
    A helper function that takes in a text string and breaks it into pieces with a maximum size, making sure all
    markdown is properly closed in the process.
    :param input_string: The input string to be split into chunks
    :param max_len: The maximum length of a given chunk
    :param splitter: The token upon which the function should try to split the string
    :return: A list of strings with closed markdown, each within `max_len` length
    """
    final_strings = []
    open_markdown = ""
    if len(input_string) <= max_len:
        return [input_string]
    while True:
        snippet = input_string[:max_len - len(open_markdown)]
        while True:
            # Find the nearest splitter character to the maximum length
            snippet = snippet.rsplit(splitter, 1)[0] or snippet
            cut_length = len(snippet)
            snippet_marked_down = open_markdown + snippet
            # Close the markdown
            snippet_marked_down, extra_md = close_markdown(snippet_marked_down)
            # If, with markdown closed, it fits, continue. Otherwise, run the loop again.
            if len(snippet_marked_down) <= max_len:
                open_markdown = extra_md
                break
            elif len(snippet.strip().rsplit(splitter, 1)) <= 1:
                snippet = snippet[:max_len - len(snippet_marked_down)]
        final_strings.append(snippet_marked_down)
        input_string = input_string[cut_length:]  # Cut down the input string for the next iteration
        if not input_string:
            return final_strings


def close_markdown(input_string):
    code_block_matches = re.findall(r"```\w+\n", input_string)
    in2 = re.sub(r"```\w+\n", "```", input_string)
    md_matches = re.findall(r"(\*\*|\*|~~|__|\|\||```|`)", in2)
    unclosed_matches = "".join({s for s in md_matches if md_matches.count(s) % 2 != 0})
    output = input_string + unclosed_matches
    if code_block_matches:
        unclosed_matches = unclosed_matches.replace("```", code_block_matches[0], 1)
    return output, unclosed_matches


def group_items(items, message="", header='```\n', footer='```', joiner='\n'):
    """
    Utility function to group a number of list items into sub-2000 length strings for posting through discord.
    Assumes every item is a string and is below 2000 symbols itself.
    :param items: list of strings to group.
    :param message: Optional message to include before the items.
    :param header: For discord formatting, defaults to putting everything into a code block.
    :param footer: The other part of the code block.
    :param joiner: The string to join items with. Called as joiner.join()
    :return: A list of joined strings.
    """
    result = []
    l_max = 2000 - len(header) + len(footer)
    l_join = len(joiner)
    l_temp = len(message)
    r_temp = []
    for i in items:
        if l_temp + len(i) + l_join > l_max:
            result.append(f"{header}{joiner.join(r_temp)}{footer}")
            r_temp = [i]
            l_temp = len(i) + l_join
        else:
            r_temp.append(i)
            l_temp += len(i) + l_join

    result.append(f"{header}{joiner.join(r_temp)}{footer}")
    if message:
        result[0] = message + result[0]

    return result


def ordinal(n):
    """
    Black magic that turns numbers into ordinal representation (1 -> 1st)
    :param n: number to be converted
    :return: string with ordinal number
    """
    return "%d%s" % (n, "tsnrhtdd"[((n // 10) % 10 != 1) * (n % 10 < 4) * n % 10::4])


def decode_json(data):
    try:
        try:
            json_str = data.decode()
        except UnicodeDecodeError:
            try:
                json_str = data.decode(encoding="windows-1252")
            except UnicodeDecodeError:
                try:
                    json_str = data.decode(encoding="windows-1250")
                except UnicodeDecodeError:
                    raise ValueError("Unable to parse file encoding. Please use UTF-8")
        else:
            if json_str[0] != "{":
                json_str = data.decode(encoding="utf-8-sig")
        json_object = json.loads(json_str)
    except json.decoder.JSONDecodeError as e:
        raise ValueError(f"Not a valid JSON file: {e}")
    return json_object


def pretty_time(seconds):
    """
    Pretty time display function
    :param seconds: time in seconds
    :return: time in weeks, days and h:mm:ss
    """
    MINUTE_SECONDS = 60
    HOUR_SECONDS = 3600
    DAY_SECONDS = 86400
    WEEK_SECONDS = 604800

    weeks, days = divmod(int(seconds), WEEK_SECONDS)
    days, hours = divmod(days, DAY_SECONDS)
    hours, minutes = divmod(hours, HOUR_SECONDS)
    minutes, seconds = divmod(minutes, MINUTE_SECONDS)

    result_list = []
    if weeks > 1:
        result_list.append(f"{weeks} weeks")
    elif weeks == 1:
        result_list.append("1 week")

    if days > 1:
        result_list.append(f"{days} days")
    elif days == 1:
        result_list.append("1 day_seconds")

    if hours > 0:
        if minutes == seconds == 0:
            if hours > 1:
                result_list.append(f"{hours} hours")
            else:
                result_list.append("1 hour")
        else:
            result_list.append(f"{hours}:{minutes:02d}:{seconds:02d}")
    elif minutes > 0:
        if seconds == 0:
            if minutes > 1:
                result_list.append(f"{minutes} minutes")
            else:
                result_list.append("1 minute")
        else:
            result_list.append(f"{minutes}:{seconds:02d}")
    elif seconds > 1:
        result_list.append(f"{seconds:02d} seconds")
    elif seconds == 1:
        result_list.append("1 second")

    return ", ".join(result_list)


def is_positive(string):
    """
    Returns True if the string is a positive word and False if the string is a negative word
    :type string: str
    :param string: string to be judged
    :return: boolean
    """
    if string.lower() in ["off", "disable", "no", "negative", "false"]:
        return False
    elif string.lower() in ["on", "enable", "yes", "affirmative", "true"]:
        return True
    else:
        raise CommandSyntaxError(f"{string} is not valid positive/negative input. "
                                 f"Allowed inputs: off/disable/no/negative/false, "
                                 "on/enable/yes/affirmatie/true.")


def verify_embed(embed: dict):
    """
    A big ugly function to verify the embed dict as best as we can.
    Made even uglier by the verbosity I choose to include in the verification.
    :param embed:
    :return:
    """

    class EmbedDict(dict):
        # A simple class that allows the output of this command to be used in place of discord.py Embed class for
        # channel.send() purposes.
        def to_dict(self):
            return self

    def option(res: dict, key: str, target: dict, scheme: (dict, int, None)):
        """
        Verification function.
        Scheme can be a length for text fields or None for url verification.
        Otherwise, it may be a dict of fields with either lengths or None for urls.
        URLs are limited to 2048 symbols by default.
        :param res: "result" dict to put the values into
        :param key: Key to be verified
        :param target: The "embed" dict to pull values from
        :param scheme: Verification scheme
        :return: No returns, the result is added to res parameter. To prevent adding empty fields.
        """
        total_len = 0
        if key in target:
            if isinstance(scheme, dict):
                res[key] = {}
                for field in scheme:
                    if field not in target[key]:
                        continue
                    if scheme[field]:
                        if len(target[key][field]) > scheme[field]:
                            raise ValueError(f"{key}[{field}] too long. (limit {scheme[field]})")
                        total_len += len(target[key][field])
                    else:
                        # verify URL
                        if len(target[key][field]) > 2048:
                            raise ValueError(f"{key}[{field}] tool long. (limit 2048)")
                        url = urlparse(target[key][field])
                        if not (url.scheme and url.netloc and url.path):
                            raise ValueError(f"{key}[{field}] invalid url.")
                    res[key][field] = target[key][field]
            else:
                if scheme:
                    if len(target[key]) > scheme:
                        raise ValueError(f"{key} too long. (limit {scheme})")
                    total_len += len(target[key])
                else:
                    if len(target[key]) > 2048:
                        raise ValueError(f"{key} too long. (limit 2048)")
                    url = urlparse(target[key])
                    if not (url.scheme and url.netloc and url.path):
                        raise ValueError(f"{key} invalid url.")
                res[key] = target[key]
        return total_len

    result = EmbedDict(type='rich')
    embed_len = 0

    embed_len += option(result, 'title', embed, 256)
    embed_len += option(result, 'description', embed, 2048)
    embed_len += option(result, 'url', embed, None)
    embed_len += option(result, 'image', embed, {"url": None})
    embed_len += option(result, 'thumbnail', embed, {"url": None})
    embed_len += option(result, 'footer', embed, {"text": 2048, "icon_url": None})
    embed_len += option(result, 'author', embed, {"name": 256, "url": None, "icon_url": None})

    if 'color' in embed:
        try:
            result['color'] = int(embed['color'], 0)
        except TypeError:
            try:
                result['color'] = int(embed['color'])
            except ValueError:
                raise ValueError("Invalid color value.")

    if 'fields' in embed:
        result['fields'] = []
        for i, field in enumerate(embed['fields']):
            if not isinstance(field.get('inline', False), bool):
                raise ValueError(f"Field {i+1}, \"inline\" must be true or false.")
            if len(str(field['name'])) > 256:
                raise ValueError(f"Field {i+1} \"name\" too long. (Limit 256)")
            if len(str(field['value'])) > 1024:
                raise ValueError(f"Field {i+1} \"value\" too long. (Limit 1024)")
            embed_len += len(str(field['name'])) + len(str(field['value']))
            result['fields'].append({
                'name': str(field['name']), 'value': str(field['value']),
                'inline': field.get('inline', False)
            })

    if embed_len > 6000:
        raise ValueError("Embed size exceed maximum size of 6000.")

    return result



from red_star.plugin_manager import BasePlugin
from red_star.command_dispatcher import Command
from red_star.rs_utils import respond, RSArgumentParser, group_items
from red_star.rs_errors import CommandSyntaxError, ChannelNotFoundError
import datetime
import shlex
import re
from discord import Message, HTTPException
from dataclasses import dataclass
from asyncio import create_task

recurDecode = {
    "h": ('hour', 'hours'),
    "d": ('day', 'days'),
    "m": ('month', 'months'),
    "y": ('year', 'years')
}


class ReminderPlugin(BasePlugin):
    name = "reminder"
    version = "1.2"
    author = "GTG3000"
    description = "A plugin for setting messages that the bot will send you at a configurable time."
    channel_types = {"reminders"}

    run_timer = False
    timer = None
    reminder_file_path = None

    # Searches for a DD/MM/YYYY@hh:mm:ss pattern.
    # All the options are optional (searching for 0 to 2/4 digits), and are in named groups for ease of use of results.
    # It *has* to be formatted // for date and :: for time though.
    # The bit between the two just has to be non-numeric, but since shlex kills whitespace it also has to be a thing.
    pattern = re.compile(
        r"(?:(?P<D>\d{0,2})/(?P<M>\d{0,2})/(?P<Y>\d{0,4}))?[^\d:/]*(?:(?P<h>\d{0,2}):(?P<m>\d{0,2}):(?P<s>\d{0,2}))?")

    @dataclass
    class Reminder:
        uid: int
        cid: int
        time: datetime.datetime
        text: str
        dm: bool
        recurring: list

        # number of days in each months, for checking
        _mdays = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)

        def __post_init__(self):
            if isinstance(self.time, str):
                self.time = datetime.datetime.fromisoformat(self.time)

        def as_dict(self):
            return {"__classhint__": "reminder",
                    "reminder": (self.uid, self.cid, self.time.isoformat(), self.text, self.dm, self.recurring)}

        def get_recurring(self):
            if self.recurring:
                if self.recurring[0] == 'y':
                    return datetime.datetime(year=self.time.year + self.recurring[1],
                                             month=self.time.month,
                                             day=self.time.day,
                                             hour=self.time.hour,
                                             minute=self.time.minute,
                                             second=self.time.second)
                elif self.recurring[0] == 'm':
                    month = self.time.month + self.recurring[1] - 1
                    year = self.time.year + month // 12
                    month = month % 12 + 1
                    day = min(self.time.day, self._mdays[month - 1])
                    return datetime.datetime(year=year,
                                             month=month,
                                             day=day,
                                             hour=self.time.hour,
                                             minute=self.time.minute,
                                             second=self.time.second)
                elif self.recurring[0] == 'd':
                    return self.time + datetime.timedelta(days=self.recurring[1])
                else:
                    return self.time + datetime.timedelta(hours=self.recurring[1])
            else:
                return False

    async def activate(self):
        def _load(cls: ReminderPlugin, obj: dict):
            if obj.pop('__classhint__', None) == 'reminder':
                return cls.Reminder(*obj['reminder'])
            else:
                return obj

        self.storage = self.config_manager.get_plugin_config_file(
                "reminders.json", json_load_args={'object_hook': lambda o: _load(self, o)},
                json_save_args={'default': lambda x: x.as_dict()})
        for guild in self.client.guilds:
            if str(guild.id) not in self.storage:
                self.storage[str(guild.id)] = []

    @Command("Remind",
             syntax="(message) [-d/--delay DD//@HH:MM:SS] [-t/--time DD/MM/YYYY@HH:MM:SS] [-p/--private] ["
                    "-r/--recurring y/m/d###]",
             doc="Store a reminder for the future. Message needs to come before the time options.\n"
                 "-d/--delay    : remind after a set delay. Does NOT support months or years due to varying time.\n"
                 "-t/--time     : remind on a given date. Date must be in the future.\n"
                 "-r/--recurring: repeat the reminder every day/month/year, format is d/m/y## with ## being an "
                 "integer number.\n"
                 "-e/--everyone : mention everyone with the reminder. Requires user having the permission.\n"
                 "-h/--here     : mention here with the reminder. Requires user having the permission.\n"
                 "Time format   : DD/MM/YYYY@hh:mm:ss. Numbers may be skipped, but a date needs two slashes and a "
                 "time needs two colons, with some non-whitespace symbol separating the two.\n"
                 "Valid input includes '23//', ':30:' and '1/1/@12::'",
             category="reminder",
             run_anywhere=True)
    async def _remind(self, msg: Message):
        parser = RSArgumentParser()
        parser.add_argument("command")
        parser.add_argument("reminder", default=[], nargs="*")
        parser.add_argument("-d", "--delay", default='')
        parser.add_argument("-t", "--time", default='')
        parser.add_argument("-p", "--private", action='store_true')
        parser.add_argument("-e", "--everyone", action='store_true')
        parser.add_argument("-h", "--here", action='store_true')
        parser.add_argument("-r", "--recurring", default='')

        try:
            args = shlex.split(msg.content)
        except ValueError as e:
            self.logger.warning("Unable to split {data.content}. {e}")
            raise CommandSyntaxError(e)

        args = parser.parse_args(args)

        if not (args['time'] or args['delay']):
            raise CommandSyntaxError("Expected a reminder time")
        else:
            time = args['time'] if args['time'] else args['delay']

        utcnow = datetime.datetime.utcnow()

        # just a way to default skipped values to "today"
        default_time = {
            'D': 0 if args['delay'] else utcnow.day,
            'M': utcnow.month,
            'Y': utcnow.year,
            'h': 0,
            'm': 0,
            's': 0
        }

        time = {k: int(v or default_time[k]) for k, v in self.pattern.match(time).groupdict().items()}

        if args['private']:
            args['everyone'] = args['here'] = False
            try:
                await msg.delete()
            except HTTPException:
                pass

        # just because the bot can mention everyone, doesn't mean that anyone with command access should be able to.
        if msg.author.permissions_in(msg.channel).mention_everyone and (args['everyone'] or args['here']):
            args['reminder'].insert(0, "@everyone:" if args['everyone'] else "@here:")

        if args['recurring'] and args['recurring'][0].lower() in 'ymdh':
            try:
                _recur = (args['recurring'][0].lower(), int(args['recurring'][1:]))
                _recurstr = f" Recurring every {recurDecode[_recur[0]][0] if _recur[1]==1 else str(_recur[1])+' '+recurDecode[_recur[0]][1]}."
                if _recur[1] <= 0:
                    raise ValueError
            except ValueError:
                raise CommandSyntaxError('invalid recurring time format')
        else:
            _recur = []
            _recurstr = ""

        try:
            if args['time']:
                time = datetime.datetime(max(time['Y'], 1), max(time['M'], 1), max(time['D'], 1),
                                         time['h'], time['m'], time['s'])
            else:
                time = utcnow + datetime.timedelta(time['D'], hours=time['h'], minutes=time['m'], seconds=time['s'])
        except ValueError as e:
            raise CommandSyntaxError(e)

        if time < utcnow:
            raise CommandSyntaxError("Red Star cannot presently alter the past.")

        self.storage[str(msg.guild.id)].append(self.Reminder(msg.author.id, msg.channel.id, time,
                                                             ' '.join(args['reminder']),
                                                             args['private'], _recur))
        self.storage.save()

        await respond(msg, f"**AFFIRMATIVE: reminder set for {time.strftime('%Y-%m-%d @ %H:%M:%S')} UTC.{_recurstr}**")

    @Command("RemindList",
             syntax="[-/del/delete index]",
             doc="Prints out a list of all reminders of the user.\nUse del argument and an index from said list to "
                 "remove reminders.",
             category="reminder")
    async def _remindlist(self, msg: Message):
        uid = msg.author.id
        gid = str(msg.guild.id)
        args = msg.clean_content.split(None, 2)
        reminder_list = [r for r in self.storage[gid] if r.uid == uid]

        if len(args) == 1:
            for split_msg in group_items((f"{i:2}|{r.time.strftime('%Y-%m-%d @ %H:%M:%S')} : {r.text:50} in "
                                          f"{'Direct Messages' if r.dm else msg.guild.get_channel(r.cid)}"
                                          for i, r in enumerate(reminder_list)),
                                         message="**Following reminders found:**"):
                await respond(msg, split_msg)
        elif len(args) == 3 and args[1].lower() in ("del", "-", "delete"):
            try:
                index = int(args[2])
            except ValueError:
                raise CommandSyntaxError("Non-integer index provided")
            if 0 <= index < len(reminder_list):
                self.storage[gid].remove(reminder_list[index])
                await respond(msg, "**AFFIRMATIVE. Reminder removed.**")
            else:
                raise CommandSyntaxError("Index out of range")
        else:
            raise CommandSyntaxError("No arguments or del/-/delete index required")

    @Command("RemindListAll",
             syntax="[-/del/delete index]",
             doc="Prints out a list of all reminders.\nUse del argument and an index from said list to "
                 "remove reminders.",
             category="reminder",
             perms={"manage_messages"})
    async def _remindlistall(self, msg: Message):
        gid = str(msg.guild.id)
        args = msg.clean_content.split(None, 2)

        users = {r.uid: str(msg.guild.get_member(r.uid)) for r in self.storage[gid]}

        if len(args) == 1:
            reminders = (f"{i:2}|{users[r.uid]:^32}|{r.time.strftime('%Y-%m-%d @ %H:%M:%S')} : "
                         f"{r.text[:50]:50} in {'Direct Messages' if r.dm else msg.guild.get_channel(r.cid)}"
                         for i, r in enumerate(self.storage[gid]))
            for split_msg in group_items(reminders, message="**Following reminders found:**"):
                await respond(msg, split_msg)
        if len(args) == 3 and args[1].lower() in ("del", "-", "delete"):
            try:
                index = int(args[2])
            except ValueError:
                raise CommandSyntaxError("Non-integer syntax provided")
            if 0 <= index < len(self.storage[gid]):
                del self.storage[gid][index]
                await respond(msg, "**AFFIRMATIVE. Reminder removed.**")
            else:
                raise CommandSyntaxError("Index out of range")
        else:
            raise CommandSyntaxError

    async def on_global_tick(self, now, _):
        save_flag = False
        for guild in filter(lambda x: str(x.id) in self.storage, self.client.guilds):
            gid = str(guild.id)
            for reminder in filter(lambda x: x.time <= now, self.storage[gid]):
                save_flag = True
                try:
                    channel = None if reminder.dm else self.channel_manager.get_channel(guild, "reminders")
                except ChannelNotFoundError:
                    channel = guild.get_channel(reminder.cid)

                if channel:
                    create_task(channel.send(f"**<@{reminder.uid}>:**\n{reminder.text}"))
                else:
                    usr = guild.get_member(reminder.uid)
                    if usr:
                        create_task(usr.send(f"**Reminder from {guild}:**\n{reminder.text}"))

                if reminder.recurring:
                    self.storage[gid].append(self.Reminder(reminder.uid, reminder.cid, reminder.get_recurring(),
                                                           reminder.text, reminder.dm, reminder.recurring))
            # it's easier to create a new list where everything fits rather than delete from old
            self.storage[gid] = [x for x in self.storage[gid] if x.time > now]
        if save_flag:
            self.storage.save()



from sys import argv
import math
import random


def _parse_rpn(args):
    args = [x for x in args]
    if len(args) == 0:
        raise SyntaxError("<rpn> tag requires arguments.")
    stack = []
    out = []

    def _dup(x):
        stack.append(x)
        stack.append(x)

    def _swap(x, y):
        stack.append(x)
        stack.append(y)

    def _modf(x):
        v, v1 = math.modf(x)
        stack.append(v)
        stack.append(v1)

    binary_ops = {
        "+": lambda x, y: stack.append(x + y),
        "-": lambda x, y: stack.append(y - x),
        "*": lambda x, y: stack.append(x * y),
        "/": lambda x, y: stack.append(y / x),
        "^": lambda x, y: stack.append(y ** x),
        "%": lambda x, y: stack.append(y % x),
        "//": lambda x, y: stack.append(y // x),
        "log": lambda x, y: stack.append(math.log(y, x)),
        "atan2": lambda x, y: stack.append(math.atan2(y, x)),
        "swap": _swap,
        "min": lambda x, y: stack.append(min(x, y)),
        "max": lambda x, y: stack.append(max(x, y)),
    }
    unary_ops = {
        "sin": lambda x: stack.append(math.sin(x)),
        "cos": lambda x: stack.append(math.cos(x)),
        "tan": lambda x: stack.append(math.tan(x)),
        "ln": lambda x: stack.append(math.log(x)),
        "pop": lambda x: out.append(x),
        "int": lambda x: stack.append(int(x)),
        "dup": _dup,
        "drop": lambda x: x,
        "modf": _modf,
        "round": lambda x: stack.append(round(x)),
        "rndint": lambda x: stack.append(random.randint(0, x))
    }
    constants = {
        "e": lambda: stack.append(math.e),
        "pi": lambda: stack.append(math.pi),
        "tau": lambda: stack.append(math.tau),
        "m2f": lambda: stack.append(3.280839895),
        "m2i": lambda: stack.append(39.37007874),
        "rnd": lambda: stack.append(random.random())
    }
    for arg in args:
        try:
            value = int(arg, 0)
        except ValueError:
            try:
                value = float(arg)
            except ValueError:
                if arg in binary_ops and len(stack) > 1:
                    binary_ops[arg](stack.pop(), stack.pop())
                elif arg in unary_ops and len(stack) >= 1:
                    unary_ops[arg](stack.pop())
                elif arg in constants:
                    constants[arg]()
            else:
                stack.append(value)
        else:
            stack.append(value)
    return [*out, *stack]


if __name__ == "__main__":
    try:
        print(" ".join(str(x) for x in _parse_rpn(argv[1:])))
    except (ValueError, ZeroDivisionError, SyntaxError) as e:
        print(e)
        raise Exception()



# Python Lisp parser

import math
import operator as op
import re
import random
import datetime
from time import time
from collections import OrderedDict
from red_star.rs_errors import CustomCommandSyntaxError
from functools import reduce

Symbol = str
Number = (int, float)
List = list

_args = 'args'
_quote = 'quote'
_if = 'if'
_set = ':='
_define = 'define'
_def = 'def'
_lambda = 'lambda'
_begin = 'do'
_unquote = 'unquote'
_access = '>>'
_while = 'while'
_print = 'print'
_try = 'try'
_raise = 'raise'

# temporarily swap a bunch of escaped character into some temporarily values and into un-escaped versions
escapes = (("\\\\", "\uff00", "\\"), ("\\\"", "\uff01", "\""), ("\\n", "\uff02", "\n"), ("\\;", "\uff03", ";"))
# search for lisp instances -
# > comments - ;.*?(?:\n|$) - start with ;, end at line end/end of file, whichever's shorter.
# > strings  - \".*?\"      - start and end with quotes, gets the shortest option.
# > brackets - \(|\)        - get closing and opening bracket symbols as separate tokens.
# > tokens   - [^()\";\S]   - get everything that isn't a special char or whitespace.
tokenizer = re.compile(r";.*?(?:\n|$)|\".*?\"|\(|\)|[^()\";\s]+", re.DOTALL)


class Empty:
    pass


def l_escape(string: str) -> str:
    for p, r, _ in escapes:
        string = string.replace(p, r)
    return string


def l_restore(string: str) -> str:
    for _, p, r in escapes:
        string = string.replace(p, r)
    return string


def l_revert(string: str) -> str:
    for r, _, p in escapes:
        string = string.replace(p, r)
    return string


def tokenize(string: str) -> list:
    return tokenizer.findall(l_escape(string))


def parse(program: str):
    return read_from_tokens(tokenize(program))


def joiner(iterable):
    return reduce(lambda s, i: s + i if i[0] in '()"' or s[-1] in '()"' else s + ' ' + i, iterable)


def reprint(ast: [list, int, str, float]) -> str:
    """
    Takes an RSLisp Abstract Syntax Tree and prints out the corresponding lisp code.
    :param ast:
    :return:
    """
    if isinstance(ast, list):
        if len(ast) > 0:
            if ast[0] == 'quote' and isinstance(ast[1], str):
                return f'"{l_revert(ast[1])}"'
            else:
                return '('+joiner([reprint(x) for x in ast])+')'
        else:
            return "()"
    else:
        return l_revert(str(ast))


def minify(program: [str, list]):
    """
    Takes a program or an abstract syntax tree and returns the corresponding lisp code, with comments and extra
    whitespace stripped.
    :param program:
    :return:
    """
    return reprint(program if isinstance(program, list) else parse(program))


def read_from_tokens(tokens):
    if len(tokens) == 0:
        raise CustomCommandSyntaxError('unexpected EOF while reading')
    token = tokens.pop(0)

    if token[0] == token[-1] == '"':
        return ['quote', l_restore(token[1:-1])]

    elif '(' == token:
        token_list = []
        while tokens[0] != ')':
            t = read_from_tokens(tokens)
            if not isinstance(t, Empty) and t != '':
                token_list.append(t)
        tokens.pop(0)
        return token_list

    elif token.startswith(';'):
        return Empty()

    elif ')' == token:
        raise CustomCommandSyntaxError('unexpected )')
    else:
        # regexp strips leading/trailing whitespace
        return atom(re.sub(r"^\s+|\s+$", "", token))


def atom(token: str):
    try:
        return int(token, 0)
    except ValueError:
        try:
            return float(token)
        except ValueError:
            token = token.lower()
            if token == "true":
                return True
            elif token == "false":
                return False
            else:
                return token


# =============================================================================================================
# Evaluator

# A user-defined Scheme procedure.
class Procedure(object):
    def __init__(self, parms, body, env):
        self.parms, self.body, self.env = parms, body, env

    def __call__(self, *args):
        return lisp_eval(self.body, Env(self.parms, args, self.env))


class Env(dict):
    def __init__(self, parms=(), args=(), outer=None, max_runtime=0):
        super().__init__()
        self.update(zip(parms, args))
        self.outer = outer
        self.timestamp = time()
        self.max_runtime = max_runtime

    # Find the innermost Env where var appears.
    def find(self, var):
        if var in self:
            return self
        elif self.outer:
            return self.outer.find(var)
        else:
            raise CustomCommandSyntaxError(f'undefined var {var}')


def get_args(args: list) -> (list, dict):
    t_list = [*args]
    t_dict = OrderedDict()
    for n, a in [*enumerate(t_list)][::-1]:
        if type(a) == str and a.startswith(':'):
            try:
                t_dict[a[1:]] = t_list.pop(n + 1)
            except IndexError:
                raise CustomCommandSyntaxError(f"supplied argument {a} given without value")
            del t_list[n]
    # noinspection PyTypeChecker
    return t_list, OrderedDict(reversed(t_dict.items()))


def _str(*args):
    if not args:
        return
    elif len(args) == 1:
        return str(args[0])
    else:
        try:
            return getattr(str, args[0])(*args[1:])
        except AttributeError:
            raise CustomCommandSyntaxError(f'str does not have method {args[0]}')


def eztime(*args):
    now = datetime.datetime.utcnow()

    if args:
        strf = args[0] if args[0] else "%Y-%m-%d @ %H:%M:%S"
        if len(args) > 1:
            # regexp searches for offset in h:m:s format.
            # each number is allowed to be negative and can pretty much stretch into forever.
            o_time = re.match(r"(?P<h>-?\d*):(?P<m>-?\d*):(?P<s>-?\d*)", args[1])
            if o_time:
                o_time = o_time.groupdict()
                delta = datetime.timedelta(hours=int(o_time['h']) if o_time['h'] else 0,
                                           minutes=int(o_time['m']) if o_time['m'] else 0,
                                           seconds=int(o_time['s']) if o_time['s'] else 0)
            else:
                raise CustomCommandSyntaxError(f"(eztime) invalid offset string \"{args[1]}\". "
                                               f"Please use H:M:S format.")
        else:
            delta = datetime.timedelta()
        now = now + delta
        return now.strftime(strf)
    else:
        return now.strftime("%Y-%m-%d @ %H:%M:%S")


def _assert(var, vartype, *opt):
    try:
        if vartype == 'int':
            return int(var, 0)
        elif vartype == 'float':
            return float(var)
        elif vartype == 'list':
            return list(*var)
    except (ValueError, TypeError):
        if opt:
            return opt[0]
        else:
            raise CustomCommandSyntaxError(f'assertion error: {var} is not a valid {vartype}')


def _sorted(iterable, *args):  # (sort iterable key reversed)
    kwargs = dict()
    if args:
        if len(args) > 0 and args[0]:
            kwargs['key'] = args[0]
        if len(args) > 1:
            kwargs['reverse'] = bool(args[1])
    return sorted(iterable, **kwargs)


def transcode(string: str, *args):
    if len(args) == 0:
        return string
    elif len(args) == 1:
        def_code = "ABCDEFGHIJKLMabcdefghijklmNOPQRSTUVWXYZnopqrstuvwxyz"
        alt_code = {
            "rot13": "NOPQRSTUVWXYZnopqrstuvwxyzABCDEFGHIJKLMabcdefghijklm",
            "circled": "ⒶⒷⒸⒹⒺⒻⒼⒽⒾⒿⓀⓁⓂⓐⓑⓒⓓⓔⓕⓖⓗⓘⓙⓚⓛⓜⓃⓄⓅⓆⓇⓈⓉⓊⓋⓌⓍⓎⓏⓝⓞⓟⓠⓡⓢⓣⓤⓥⓦⓧⓨⓩ",
            "circled_neg": "🅐🅑🅒🅓🅔🅕🅖🅗🅘🅙🅚🅛🅜🅐🅑🅒🅓🅔🅕🅖🅗🅘🅙🅚🅛🅜🅝🅞🅟🅠🅡🅢🅣🅤🅥🅦🅧🅨🅩🅝🅞🅟🅠🅡🅢🅣🅤🅥🅦🅧🅨🅩",
            "fwidth": "ＡＢＣＤＥＦＧＨＩＪＫＬＭａｂｃｄｅｆｇｈｉｊｋｌｍＮＯＰＱＲＳＴＵＶＷＸＹＺｎｏｐｑｒｓｔｕｖｗｘｙｚ",
            "mbold": "𝐀𝐁𝐂𝐃𝐄𝐅𝐆𝐇𝐈𝐉𝐊𝐋𝐌𝐚𝐛𝐜𝐝𝐞𝐟𝐠𝐡𝐢𝐣𝐤𝐥𝐦𝐍𝐎𝐏𝐐𝐑𝐒𝐓𝐔𝐕𝐖𝐗𝐘𝐙𝐧𝐨𝐩𝐪𝐫𝐬𝐭𝐮𝐯𝐰𝐱𝐲𝐳",
            "mbolditalic": "𝑨𝑩𝑪𝑫𝑬𝑭𝑮𝑯𝑰𝑱𝑲𝑳𝑴𝒂𝒃𝒄𝒅𝒆𝒇𝒈𝒉𝒊𝒋𝒌𝒍𝒎𝑵𝑶𝑷𝑸𝑹𝑺𝑻𝑼𝑽𝑾𝑿𝒀𝒁𝒏𝒐𝒑𝒒𝒓𝒔𝒕𝒖𝒗𝒘𝒙𝒚𝒛",
            "frakturbold": "𝕬𝕭𝕮𝕯𝕰𝕱𝕲𝕳𝕴𝕵𝕶𝕷𝕸𝖆𝖇𝖈𝖉𝖊𝖋𝖌𝖍𝖎𝖏𝖐𝖑𝖒𝕹𝕺𝕻𝕼𝕽𝕾𝕿𝖀𝖁𝖂𝖃𝖄𝖅𝖓𝖔𝖕𝖖𝖗𝖘𝖙𝖚𝖛𝖜𝖝𝖞𝖟",
            "fraktur": "𝔄𝔅ℭ𝔇𝔈𝔉𝔊ℌℑ𝔍𝔎𝔏𝔐𝔞𝔟𝔠𝔡𝔢𝔣𝔤𝔥𝔦𝔧𝔨𝔩𝔪𝔑𝔒𝔓𝔔ℜ𝔖𝔗𝔘𝔙𝔚𝔛𝔜ℨ𝔫𝔬𝔭𝔮𝔯𝔰𝔱𝔲𝔳𝔴𝔵𝔶𝔷",
            "scriptbold": "𝓐𝓑𝓒𝓓𝓔𝓕𝓖𝓗𝓘𝓙𝓚𝓛𝓜𝓪𝓫𝓬𝓭𝓮𝓯𝓰𝓱𝓲𝓳𝓴𝓵𝓶𝓝𝓞𝓟𝓠𝓡𝓢𝓣𝓤𝓥𝓦𝓧𝓨𝓩𝓷𝓸𝓹𝓺𝓻𝓼𝓽𝓾𝓿𝔀𝔁𝔂𝔃",
            "script": "𝒜𝐵𝒞𝒟𝐸𝐹𝒢𝐻𝐼𝒥𝒦𝐿𝑀𝒶𝒷𝒸𝒹𝑒𝒻𝑔𝒽𝒾𝒿𝓀𝓁𝓂𝒩𝒪𝒫𝒬𝑅𝒮𝒯𝒰𝒱𝒲𝒳𝒴𝒵𝓃𝑜𝓅𝓆𝓇𝓈𝓉𝓊𝓋𝓌𝓍𝓎𝓏",
            "sans": "𝖠𝖡𝖢𝖣𝖤𝖥𝖦𝖧𝖨𝖩𝖪𝖫𝖬𝖺𝖻𝖼𝖽𝖾𝖿𝗀𝗁𝗂𝗃𝗄𝗅𝗆𝖭𝖮𝖯𝖰𝖱𝖲𝖳𝖴𝖵𝖶𝖷𝖸𝖹𝗇𝗈𝗉𝗊𝗋𝗌𝗍𝗎𝗏𝗐𝗑𝗒𝗓",
            "sansbold": "𝗔𝗕𝗖𝗗𝗘𝗙𝗚𝗛𝗜𝗝𝗞𝗟𝗠𝗮𝗯𝗰𝗱𝗲𝗳𝗴𝗵𝗶𝗷𝗸𝗹𝗺𝗡𝗢𝗣𝗤𝗥𝗦𝗧𝗨𝗩𝗪𝗫𝗬𝗭𝗻𝗼𝗽𝗾𝗿𝘀𝘁𝘂𝘃𝘄𝘅𝘆𝘇",
            "sansbolditalic": "𝘼𝘽𝘾𝘿𝙀𝙁𝙂𝙃𝙄𝙅𝙆𝙇𝙈𝙖𝙗𝙘𝙙𝙚𝙛𝙜𝙝𝙞𝙟𝙠𝙡𝙢𝙉𝙊𝙋𝙌𝙍𝙎𝙏𝙐𝙑𝙒𝙓𝙔𝙕𝙣𝙤𝙥𝙦𝙧𝙨𝙩𝙪𝙫𝙬𝙭𝙮𝙯",
            "sansitalic": "𝘈𝘉𝘊𝘋𝘌𝘍𝘎𝘏𝘐𝘑𝘒𝘓𝘔𝘢𝘣𝘤𝘥𝘦𝘧𝘨𝘩𝘪𝘫𝘬𝘭𝘮𝘕𝘖𝘗𝘘𝘙𝘚𝘛𝘜𝘝𝘞𝘟𝘠𝘡𝘯𝘰𝘱𝘲𝘳𝘴𝘵𝘶𝘷𝘸𝘹𝘺𝘻",
            "parenthesized": "⒜⒝⒞⒟⒠⒡⒢⒣⒤⒥⒦⒧⒨⒜⒝⒞⒟⒠⒡⒢⒣⒤⒥⒦⒧⒨⒩⒪⒫⒬⒭⒮⒯⒰⒱⒲⒳⒴⒵⒩⒪⒫⒬⒭⒮⒯⒰⒱⒲⒳⒴⒵",
            "doublestruck": "𝔸𝔹ℂ𝔻𝔼𝔽𝔾ℍ𝕀𝕁𝕂𝕃𝕄𝕒𝕓𝕔𝕕𝕖𝕗𝕘𝕙𝕚𝕛𝕜𝕝𝕞ℕ𝕆ℙℚℝ𝕊𝕋𝕌𝕍𝕎𝕏𝕐ℤ𝕟𝕠𝕡𝕢𝕣𝕤𝕥𝕦𝕧𝕨𝕩𝕪𝕫",
            "region": "🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇳🇴🇵🇶🇷🇸🇹🇺🇻🇼🇽🇾🇿🇳🇴🇵🇶🇷🇸🇹🇺🇻🇼🇽🇾🇿",
            "squared": "🄰🄱🄲🄳🄴🄵🄶🄷🄸🄹🄺🄻🄼🄰🄱🄲🄳🄴🄵🄶🄷🄸🄹🄺🄻🄼🄽🄾🄿🅀🅁🅂🅃🅄🅅🅆🅇🅈🅉🄽🄾🄿🅀🅁🅂🅃🅄🅅🅆🅇🅈🅉",
            "squared_neg": "🅰🅱🅲🅳🅴🅵🅶🅷🅸🅹🅺🅻🅼🅰🅱🅲🅳🅴🅵🅶🅷🅸🅹🅺🅻🅼🅽🅾🅿🆀🆁🆂🆃🆄🆅🆆🆇🆈🆉🅽🅾🅿🆀🆁🆂🆃🆄🆅🆆🆇🆈🆉",
            "subscript": "ₐBCDₑFGₕᵢⱼₖₗₘₐbcdₑfgₕᵢⱼₖₗₘₙₒₚQᵣₛₜᵤᵥWₓYZₙₒₚqᵣₛₜᵤᵥwₓyz",
            "superscript": "ᴬᴮᶜᴰᴱᶠᴳᴴᴵᴶᴷᴸᴹᵃᵇᶜᵈᵉᶠᵍʰⁱʲᵏˡᵐᴺᴼᴾQᴿˢᵀᵁⱽᵂˣʸᶻⁿᵒᵖqʳˢᵗᵘᵛʷˣʸᶻ",
            "inverted": "ɐqɔpǝɟƃɥıɾʞןɯɐqɔpǝɟƃɥıɾʞןɯuodbɹsʇn𐌡ʍxʎzuodbɹsʇnʌʍxʎz",
            "reversed": "AdↃbƎꟻGHIJK⅃MAdↄbɘꟻgHijklmᴎOꟼpᴙꙄTUVWXYZᴎoqpᴙꙅTUvwxYz",
            "smallcaps": "ABCDEFGHIJKLMᴀʙᴄᴅᴇꜰɢʜɪᴊᴋʟᴍNOPQRSTUVWXYZɴᴏᴩqʀꜱᴛᴜᴠᴡxyᴢ",
            "weird1": "ልጌርዕቿቻኗዘጎጋጕረጠልጌርዕቿቻኗዘጎጋጕረጠክዐየዒዪነፕሁሀሠሸሃጊክዐየዒዪነፕሁሀሠሸሃጊ",
            "weird2": "ДБҀↁЄFБНІЈЌLМаъсↁэfБЂіјкlмИФРQЯЅГЦVЩЖЧZиорqѓѕтцvшхЎz",
            "weird3": "ค๒ƈɗﻉिﻭɦٱﻝᛕɭ๓ค๒ƈɗﻉिﻭɦٱﻝᛕɭ๓กѻρ۹ɼรՇપ۷ฝซץչกѻρ۹ɼรՇપ۷ฝซץչ",
            "weird4": "αв¢∂єƒﻭнιנкℓмαв¢∂єƒﻭнιנкℓмησρ۹яѕтυνωχуչησρ۹яѕтυνωχуչ",
            "weird5": "ค๒ς๔єŦﻮђเןкɭ๓ค๒ς๔єŦﻮђเןкɭ๓ภ๏קợгรՇยשฬאץչภ๏קợгรՇยשฬאץչ",
            "weird6": "ﾑ乃cd乇ｷgんﾉﾌズﾚﾶﾑ乃cd乇ｷgんﾉﾌズﾚﾶ刀oｱq尺丂ｲu√wﾒﾘ乙刀oｱq尺丂ｲu√wﾒﾘ乙",
            "sbancient": ""
        }
        if args[0].lower() == 'help':
            return "```\nAVAILABLE TRANSCODINGS:\n" + "\n".join(alt_code.keys()) + "```"
        elif args[0].lower() not in alt_code:
            raise CustomCommandSyntaxError(f"{args[0].lower()} is not a supported transcoding. Use (transcode "
                                           f"\"help\") to get a list of available transcodings.")
        else:
            alt_code = alt_code[args[0].lower()]
    else:
        def_code = args[0]
        alt_code = args[1]
        if len(def_code) != len(alt_code):
            raise CustomCommandSyntaxError("transcode: To and From transcoding patterns must be the same length.")
    return string.translate(str.maketrans(def_code, alt_code))


def standard_env(*_, **kwargs):
    env = Env(**kwargs)
    env.update(vars(math))
    env.update({
        '+': op.add,
        '-': lambda *x: op.sub(*x) if len(x) > 1 else -x[0],
        '*': op.mul, '/': op.truediv, '//': op.floordiv, '%': op.mod, '**': op.pow,
        '>': op.gt, '<': op.lt, '>=': op.ge, '<=': op.le, '==': op.eq, '<>': op.xor,
        '!=': op.ne,
        '#': lambda x, y: y[x],
        'abs': abs,
        'append': lambda x, y: x.append(y) if type(x) == list else x + y,
        'apply': lambda proc, args: proc(*args),
        'do': lambda *x: x[-1],
        'car': lambda x: x[0],
        'cdr': lambda x: x[1:],
        'cons': lambda x, y: [x] + y,
        'is': op.is_,
        'in': op.contains,
        'len': len,
        'list': lambda *x: list(x),
        'l': lambda *x: list(x),
        'tolist': list,
        '2l': list,
        'slice': slice,
        'range': range,
        'list?': lambda x: isinstance(x, list),
        'map': lambda *x: list(map(*x)),
        'imap': map,
        'sum': sum,
        'max': max,
        'min': min,
        'all': all,
        'any': any,
        'filter': filter,
        'reduce': reduce,
        'sort': _sorted,
        'reverse': lambda x: x[::-1],
        'ireverse': reversed,
        'pass': lambda *x: None,
        'not': op.not_,
        'and': op.and_,
        'or': op.or_,
        'null?': lambda x: x == [],
        'number?': lambda x: isinstance(x, Number),
        'procedure?': callable,
        'round': round,
        'symbol?': lambda x: isinstance(x, Symbol),
        'assert': _assert,
        'f': lambda *x: "".join(map(str, x)),

        'chr': chr,
        'ord': ord,

        'int': int,
        'float': float,
        'dict': dict,
        'zip': zip,

        'resub': re.sub,
        'rematch': re.match,
        'refindall': re.findall,

        'str': _str,
        'transcode': transcode,

        'random': random.random,
        'randint': random.randint,
        'choice': lambda *x: random.choices(*x).pop(),

        'eztime': eztime,
        'time': time,
        'ezchoice': lambda *x: random.choice(x),

        # to be overriden by the cc function
        "username": "",
        "usernick": "",
        "usermention": "",
        "authorname": "",
        "authornick": "",
        "argstring": "",
        "args": [],

        "_rsoutput": ""
    })
    return env


# recursively access the list and set the last item
def _lset(lst, val, *indexes):
    if len(indexes) == 1:
        lst[indexes[0]] = val
    else:
        _lset(lst[indexes[0]], val, *indexes[1:])


# search for a list item recursively
def _lget(lst, *indexes):
    if len(indexes) == 0:
        return lst
    elif len(indexes) == 1:
        return lst[indexes[0]]
    else:
        return _lget(lst[indexes[0]], *indexes[1:])


def isnum(test):
    try:
        int(test)
        return True
    except ValueError:
        return False


# Evaluate an expression in an environment.
# noinspection PyDefaultArgument
def lisp_eval(x, env=None):
    if env is None:
        env = standard_env()
    if env.max_runtime != 0 and time() - env.timestamp > env.max_runtime:
        raise CustomCommandSyntaxError("The command ran too long.")
    try:
        if isinstance(x, Empty):
            return
        elif isinstance(x, Symbol):  # variable reference
            l, *ind = x.split(':')
            ind = [int(x) if isnum(x) else lisp_eval(x, env) for x in ind]
            return _lget(env.find(l)[l], *ind)
        elif not isinstance(x, list):  # constant literal
            return x
        elif x[0] == _args:
            argstring = env.find('argstring')['argstring']
            arglist = env.find('args')['args']
            if len(x) == 1:
                return argstring
            if len(x) == 2:
                if x[1] == '*':
                    return arglist
                else:
                    try:
                        return arglist[x[1]]
                    except IndexError:
                        return None
            else:
                try:
                    if x[1] == '*':
                        return arglist[:x[2]]
                    elif x[2] == '*':
                        return arglist[x[1]:]
                    else:
                        return arglist[x[1]:x[2]]
                except IndexError:
                    return []
        elif x[0] == _quote:  # quotation
            (_, exp) = x
            return exp
        elif x[0] == _access:  # attempt using an object method
            a = list(map(lambda i: lisp_eval(i, env), x[1:]))
            try:
                ar, kw = get_args(a[2:])
                return getattr(a[1], a[0])(*ar, **kw)
            except AttributeError:
                raise CustomCommandSyntaxError(f'{type(a[1])} has no method {a[0]}')
        elif x[0] == _if:  # conditional (if bool then else)
            (_, test, conseq, alt) = x
            exp = (conseq if lisp_eval(test, env) else alt)
            return lisp_eval(exp, env)
        elif x[0] in [_define, _def]:  # definition
            (_, var, exp) = x
            env[var] = lisp_eval(exp, env)
        elif x[0] == _set:  # assignment
            (_, var, exp) = x
            if ':' in var:
                l, *ind = var.split(':')
                ind = [int(x) if isnum(x) else lisp_eval(x, env) for x in ind]
                _lset(env.find(l)[l], lisp_eval(exp, env), *ind)
            else:
                env.find(var)[var] = lisp_eval(exp, env)
        elif x[0] == _lambda:  # procedure
            (_, parms, body) = x
            return Procedure(parms, body, env)
        elif x[0] == _while:  # while loop
            while lisp_eval(x[1], env):
                lisp_eval(x[2], env)
        elif x[0] == _print:  # prints into "_rsoutput" variable
            try:
                env.find('_rsoutput')['_rsoutput'] += f'{" ".join(map(lambda y: str(lisp_eval(y,env)), x[1:]))}\n'
            except IndexError:
                env.find('_rsoutput')['_rsoutput'] += '\n'
        elif x[0] == _try:  # (try (body) (except)) - returns result of body if successful or evaluates except if not
            expr, *args = x[1:]
            try:
                return lisp_eval(expr, env)
            except Exception as e:
                if len(args) >= 1:
                    return lisp_eval(args[0], env)
                else:
                    return e
        elif x[0] == _unquote:
            return lisp_eval(lisp_eval(x[1], env), env)
        elif x[0] == _raise:
            raise CustomCommandSyntaxError(lisp_eval(x[1], env))
        else:  # procedure call
            proc = lisp_eval(x[0], env)
            args = [lisp_eval(arg, env) for arg in x[1:]]
            return proc(*args)
    except Exception as e:
        if len(str(e)) > 1500:
            e = "..." + re.match(r"(?:.+)(\(.+?\): .+?$)", str(e)).group(1)
        try:
            raise CustomCommandSyntaxError(f"({x[0]}): {e}")
        except IndexError:
            raise CustomCommandSyntaxError(e)



import asyncio
import logging
from argparse import ArgumentParser
from discord.errors import LoginFailure
from logging.handlers import RotatingFileHandler
from os import chdir
from pathlib import Path
from red_star.client import RedStar


def main():
    chdir(Path(__file__).parent)
    default_user_dir = Path.home() / ".red_star"

    parser = ArgumentParser(
        description="General-purpose Discord bot with administration and entertainment functions.")
    parser.add_argument("-v", "--verbose", "--debug", action="count", default=0,
                        help="Enables debug output. Calling multiple times increases verbosity; two calls enables "
                             "discord.py debug output, and three calls enables asyncio's debug mode.")
    conf_path_group = parser.add_mutually_exclusive_group()
    conf_path_group.add_argument("-d", "--directory", type=Path, default=default_user_dir,
                                 help="Sets the directory in which configs, logs, and data will be stored.")
    conf_path_group.add_argument("-p", "--portable", action="store_true",
                                 help="Runs Red Star in portable mode. In portable mode, data files will be stored "
                                      "in the installation directory.")
    parser.add_argument("-l", "--logfile", type=str, default="red_star.log", help="Sets the name of the log file.")
    args = parser.parse_args()

    if args.verbose > 0:
        loglevel = logging.DEBUG
        logging.captureWarnings(True)
    else:
        loglevel = logging.INFO

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s # %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    base_logger = logging.getLogger()
    stream_logger = logging.StreamHandler()
    stream_logger.setLevel(loglevel)
    stream_logger.setFormatter(formatter)
    base_logger.addHandler(stream_logger)

    storage_dir = Path.cwd() if args.portable else args.directory

    if not storage_dir.exists():
        base_logger.warning(f"Specified config directory {storage_dir} does not exist! Creating now...")
        config_folder = storage_dir / "config"
        config_folder.mkdir(parents=True)
        plugins_folder = storage_dir / "plugins"
        plugins_folder.mkdir(parents=True)
        plugin_init = plugins_folder / "__init__.py"
        plugin_init.touch()

    logfile = storage_dir / args.logfile
    if not logfile.exists():
        logfile.touch()

    file_logger = RotatingFileHandler(logfile, maxBytes=1048576, backupCount=5, encoding="utf-8")
    file_logger.setLevel(loglevel)
    file_logger.setFormatter(formatter)
    base_logger.addHandler(file_logger)

    loop = asyncio.get_event_loop()

    if args.verbose >= 3:
        loop.set_debug(True)
        logging.getLogger("asyncio").setLevel(logging.DEBUG)
    else:
        logging.getLogger("asyncio").setLevel(logging.INFO)

    bot = RedStar(storage_dir, args)
    try:
        bot.run(bot.config["token"])
    except LoginFailure:
        base_logger.error("Bot token is invalid! Please confirm that the token exists in configuration and is correct.")
        raise SystemExit


if __name__ == "__main__":
    main()



import inspect
import logging
from asyncio import sleep
from sys import exc_info
from red_star.rs_errors import ChannelNotFoundError, CommandSyntaxError, UserPermissionError
from discord import Forbidden
from red_star.rs_utils import respond, sub_user_data


class CommandDispatcher:
    def __init__(self, client):
        self.client = client
        self.config_manager = client.config_manager
        self.logger = logging.getLogger("red_star.command_dispatcher")
        try:
            self.conf = client.config_manager.config["command_dispatcher"]
        except KeyError:
            client.config_manager.config["command_dispatcher"] = {"default": {"command_prefix": "!"}}
            self.conf = client.config_manager.config["command_dispatcher"]
        self.default_config = self.conf.get("default", {"command_prefix": "!"})

        self.commands = {}
        self.last_error = None

    def register_plugin(self, plugin):
        for _, mth in inspect.getmembers(plugin, predicate=inspect.ismethod):
            if hasattr(mth, "_command"):
                self.register(mth, mth.name.lower())

    def deregister_plugin(self, plugin):
        for _, mth in inspect.getmembers(plugin, predicate=inspect.ismethod):
            if hasattr(mth, "_command"):
                self.deregister(mth, mth.name.lower())

    def register(self, fn, name, is_alias=False):
        """
        Register commands in the command list and handle conflicts. If it's an
        alias, we don't want it to overwrite non-alias commands. Otherwise,
        overwrite based on priority, and failing that, load order.
        :param fn: The command function, with its added information.
        :param name: The command's name, for indexing.
        :param is_alias: A boolean that tells the registrar not to overwrite other commands.
        :return: None.
        """
        self.logger.debug(f"Registering command {name} from {fn.__self__.name}.")

        if name in self.commands:
            oldfn = self.commands[name]
            if fn.priority >= oldfn.priority and not is_alias:
                self.commands[name] = fn
                self.logger.warning(f"Command {name} from {fn.__self__.name} overwrites command {oldfn.name} from "
                                    f"{oldfn.__self__.name}!")
            else:
                self.logger.warning(f"Command {oldfn.name} from {oldfn.__self__.name} overwrites command {name} from "
                                    f"{fn.__self__.name}!")
        else:
            self.commands[name] = fn

        if hasattr(fn, "aliases") and not is_alias:
            for alias in fn.aliases:
                self.register(fn, alias.lower(), is_alias=True)

    def deregister(self, fn, name, is_alias=False):
        """
        Deregister commands from the command list when their plugin is deactivated.
        :param fn: The command function, with its added information.
        :param name: The command's name, for indexing.
        :param is_alias: A boolean that tells the registrar not to overwrite other commands.
        :return: None.
        """
        self.logger.debug(f"Deregistering command {name} from {fn.__self__.name}.")

        if name in self.commands:
            oldfn = self.commands[name]
            # Make sure the command isn't another plugin's
            if fn == oldfn:
                del self.commands[name]
        else:
            self.logger.debug(f"Could not deregister command {name}, no such command!")

        if hasattr(fn, "aliases") and not is_alias:
            for alias in fn.aliases:
                self.deregister(fn, alias, is_alias=True)

    # noinspection PyBroadException
    async def run_command(self, command, msg, dm_cmd=False):
        try:
            fn = self.commands[command]
        except KeyError:
            return
        if dm_cmd:
            if msg.author.id not in self.config_manager.config.get("bot_maintainers", []):
                return
            if not fn.dm_command:
                return
            try:
                await fn(msg)
            except CommandSyntaxError as e:
                err = e if e else "Invalid syntax."
                if fn.syntax:
                    await respond(msg, f"**WARNING: {err} ANALYSIS: Proper usage: {fn.name} {fn.syntax}.**")
                else:
                    await respond(msg, f"**WARNING: {err}**")
            except Exception:
                self.last_error = exc_info()
                self.logger.exception("Exception occurred in command. ", exc_info=True)
                await respond(msg, "**WARNING: Error occurred while running command.**")

        else:
            gid = str(msg.guild.id)
            try:
                if not fn.run_anywhere:
                    try:
                        cmd_channel = self.client.channel_manager.get_channel(msg.guild, "commands")
                        if msg.channel != cmd_channel:
                            return
                    except ChannelNotFoundError:
                        pass
                await fn(msg)
                if fn.delcall:
                    await sleep(1)
                    try:
                        await msg.delete()
                    except Forbidden:
                        pass
            except CommandSyntaxError as e:
                err = e if e else "Invalid syntax."
                if fn.syntax:
                    deco = self.conf[gid]["command_prefix"]
                    await respond(msg, f"**WARNING: {err} ANALYSIS: Proper usage: {deco}{fn.name} {fn.syntax}.**")
                else:
                    await respond(msg, f"**WARNING: {err}**")
            except UserPermissionError as e:
                err = f"\nANALYSIS: {e}" if str(e) else ""
                await respond(msg, sub_user_data(msg.author,
                                                 f"**NEGATIVE. INSUFFICIENT PERMISSION: <usernick>.{err}**"))
            except Forbidden:
                await respond(msg, "**NEGATIVE. This unit does not have permission to perform that action.**")
            except ChannelNotFoundError as e:
                await respond(msg, f"**NEGATIVE. Channel type `{e}` is not set on this server.**")
            except Exception:
                self.last_error = exc_info()
                self.logger.exception("Exception occurred in command. ", exc_info=True)
                await respond(msg, "**WARNING: Error occurred while running command.**")

    # Event hooks

    async def command_check(self, msg):
        try:
            gid = str(msg.guild.id)
            if gid not in self.conf:
                self.conf[gid] = self.default_config.copy()
            deco = self.conf[gid]["command_prefix"]
            dm_cmd = False
        except AttributeError:  # Oops, it's a DM isn't it
            deco = "!"
            dm_cmd = True
        if msg.author != self.client.user:
            cnt = msg.content
            if cnt.startswith(deco):
                cmd = cnt[len(deco):].split()[0].lower()
                if cmd in self.commands:
                    await self.run_command(cmd, msg, dm_cmd=dm_cmd)


class Command:
    """
    Defines a decorator that encapsulates a chat command. Provides a common
    interface for all commands, including roles, documentation, usage syntax,
    and aliases.
    """

    def __init__(self, name, *aliases, perms=None, doc=None, syntax=None, priority=0, delcall=False,
                 run_anywhere=False, bot_maintainers_only=False, dm_command=False, category="other"):
        if syntax is None:
            syntax = ()
        if isinstance(syntax, str):
            syntax = (syntax,)
        if doc is None:
            doc = ""
        self.name = name
        if not perms:
            perms = set()
        if isinstance(perms, str):
            perms = {perms}
        self.perms = perms
        self.syntax = syntax
        self.human_syntax = " ".join(syntax)
        self.doc = doc
        self.aliases = aliases
        self.priority = priority
        self.delcall = delcall
        self.run_anywhere = run_anywhere
        self.category = category
        self.bot_maintainers_only = bot_maintainers_only
        self.dm_command = dm_command

    def __call__(self, f):
        """
        Whenever a command is called, its handling gets done here.

        :param f: The function the Command decorator is wrapping.
        :return: The now-wrapped command, with all the trappings.
        """

        async def wrapped(s, msg):
            if msg.guild is None and self.dm_command:  # The permission check was handled pre-call.
                return await f(s, msg)
            user_perms = {x for x, y in msg.author.permissions_in(msg.channel) if y}
            if msg.guild.voice_client:
                user_perms |= {x for x, y in msg.author.permissions_in(msg.guild.voice_client.channel) if y}
            if (not user_perms >= self.perms or self.bot_maintainers_only) \
                    and msg.author.id not in s.config_manager.config.get("bot_maintainers", []):
                raise UserPermissionError
            return await f(s, msg)
        wrapped._command = True
        wrapped.aliases = self.aliases
        wrapped.__doc__ = self.doc
        wrapped.name = self.name
        wrapped.perms = self.perms
        wrapped.syntax = self.human_syntax
        wrapped.priority = self.priority
        wrapped.delcall = self.delcall
        wrapped.run_anywhere = self.run_anywhere
        wrapped.dm_command = self.dm_command
        wrapped.category = self.category
        return wrapped
